(*C*
      Embedding UNITY into the Calculus of Inductive Constructions
                                     Using the COQ Proof Assistent
      Author:  Mark-Oliver Stehr
               University of Hamburg
               Department of Computer Science
               Theoretical Foundations of Computer Science Group 
               stehr@informatik.uni-hamburg.de
      Date:    September 1998
*C*)


Load unity_init.


Variables f,g : nat->nat.
Axiom f_is_ascending : (m:nat)(le m (f m)).
Axiom g_is_ascending : (m:nat)(le m (g m)).
Axiom f_is_monotone : (m,n:nat)(le m n)->(le (f m) (f n)).
Axiom g_is_monotone : (m,n:nat)(le m n)->(le (g m) (g n)).


Definition common := [t:nat] (f t)=t /\ (g t)=t.


Variable cmt : nat.
Axiom cmt_exists : (common cmt).
Axiom cmt_smallest : (m:nat)(common m)->(le cmt m).



Inductive u_var : Set := u_time : u_var.


Definition u_vars := (S O).

Definition u_ith_var := [i:nat] u_time.

Theorem u_var_finite : (v:u_var)
  (Ex [i:nat] (lt i u_vars) /\ (u_ith_var i)=v).

Induction v.
Exists  O.
Split.
Unfold u_vars.
Apply lt_n_Sn.
Unfold u_ith_var.
Reflexivity.
Qed.


Theorem u_var_eq_dec : (v,v':u_var){(v=v')}+{~(v=v')}.

Induction v.
Induction v'.
Left.
Reflexivity.
Qed.


Inductive u_type_name : Set := u_nat : u_type_name.

Theorem u_type_name_eq_dec:(n,n':u_type_name){(n=n')}+{~(n=n')}.

Induction n.
Induction n'.
Left.
Reflexivity.
Qed.



Definition u_type := [n:u_type_name]
  Cases n of u_nat => nat end.


Definition u_var_type_name := [v:u_var]
  Cases v of u_time => u_nat end.


Load unity_states.


Inductive u_tr : Set :=
  update_f : u_tr | update_g : u_tr.


Lemma u_tr_eq_dec:(t,t':u_tr){(t=t')}+{~(t=t')}.

Induction t.
Induction t'.
Left.
Reflexivity.
Right.
Discriminate.
Induction t'.
Right.
Discriminate.
Left.
Reflexivity.
Qed.


Definition u_effect := [t:u_tr] 
  Cases t of 
          update_f => [s:u_st] (u_assign s u_time (f (s u_time)))
        | update_g => [s:u_st] (u_assign s u_time (g (s u_time)))
  end.


Load unity_program.

Definition u_pr := (cons update_f (cons update_g (nil u_tr))).


Load unity_logic.
Load unity_depend.
Load unity_safety.
Load unity_liveness.




Lemma CMT_S1 : (m:nat)(u_co u_pr ([s:u_st] (s u_time)=m) 
                    ([s:u_st] (le (s u_time) (f m)) \/ (le (s u_time) (g m)))).

Intros.
Unfold u_pr.
Simpl.
Split.
Intros.
Rewrite -> H.
Rewrite -> u_assign_changed.
Left.
Apply le_n.
Split.
Intros.
Rewrite -> H.
Rewrite -> u_assign_changed.
Right.
Apply le_n.
Exact I.
Qed.


Lemma CMT_S2 : (n:nat)(u_co u_pr
  [s:u_st]((common n)->(le (s u_time) n))
  (u_ex ? [m:nat] [s:u_st]((common n)->(le m n))/\
                          ((le (s u_time) (f m))\/(le (s u_time) (g m))))).

Intros.
Generalize (u_elimination u_pr u_time
             [s:u_st](common n)->(le (s u_time) n)
             [m:nat]
              [s:u_st](le (s u_time) (f m))\/(le (s u_time) (g m))).
Intro.
Cut (v:u_var)
     ~v=u_time->(u_independent [s:u_st](common n)->(le (s u_time) n) v).
Intro.
Cut (m:(u_var_type u_time))
     (u_co u_pr [s:u_st](s u_time)=m
       [s:u_st](le (s u_time) (f m))\/(le (s u_time) (g m))).
Intro.
Generalize (H H0 H1).
Clear  H.
Intro.
Apply u_co_implication_r with q:=(u_ex (u_var_type u_time)
                                   [m:(u_var_type u_time)]
                                    (u_and
                                      (u_with
                                        [s:u_st]
                                         (common n)->(le (s u_time) n)
                                        u_time m)
                                      [s:u_st]
                                       (le (s u_time) (f m))
                                       \/(le (s u_time) (g m)))).
Unfold u_everywhere u_imp u_and u_with u_var_type.
Unfold u_type u_var_type_name.
Unfold u_ex.
Intros.
Elim H2.
Intros t H3.
Exists  t.
Elim H3.
Intros.
Split.
Intro.
Generalize (H4 H6).
Rewrite -> u_assign_changed.
Intro.
Assumption.
Assumption.
Assumption.
Unfold u_var_type.
Unfold u_var_type_name u_type.
Exact CMT_S1.
Induction v.
Intro.
Apply False_ind.
Unfold not in H0.
Apply H0.
Reflexivity.
Qed.


Lemma CMT_S3 : (n:nat)(u_stable u_pr
             ([s:u_st] (common n) -> (le (s u_time) n))).

Intro.
Unfold u_stable.
Apply u_co_implication_r with q:=(u_ex ?
                                   [m:nat]
                                    [s:u_st]
                                     ((common n)->(le m n))
                                     /\((le (s u_time) (f m))
                                        \/(le (s u_time) (g m)))).
Unfold u_everywhere u_imp u_ex.
Intros.
Elim H.
Intros t H1.
Clear  H.
Elim H1.
Intros.
Clear  H1.
Generalize (H H0).
Intro.
Generalize (f_is_monotone t n H1).
Intro.
Generalize (g_is_monotone t n H1).
Intro.
Unfold common in H0.
Elim H0.
Intros.
Rewrite H5 in H3.
Rewrite H6 in H4.
Elim H2.
Intro.
Apply le_trans with (f t).
Assumption.
Assumption.
Intro.
Apply le_trans with (g t).
Assumption.
Assumption.
Apply CMT_S2.
Qed.


Lemma CMT_S4 : (n:nat)(u_ind_invariant u_pr [s:u_st](s u_time)=O
             ([s:u_st] (common n) -> (le (s u_time) n))).

Intro.
Unfold u_ind_invariant.
Split.
Unfold u_everywhere u_imp.
Intros.
Rewrite -> H.
Apply le_O_n.
Apply CMT_S3.
Qed.


Theorem CMT_S5 : (u_ind_invariant u_pr 
  [s:u_st](s u_time)=O [s:u_st](le (s u_time) cmt)).

Generalize (CMT_S4 cmt).
Intros.
Apply u_ind_invariant_subst with 3:=H.
Unfold u_everywhere u_imp.
Auto.
Unfold u_everywhere u_iff.
Intros.
Split.
Intros.
Apply H0.
Apply cmt_exists.
Intros.
Trivial.
Qed.


Theorem CMT_S6 : (m:nat)
  (u_co u_pr [s:u_st]m=(s u_time) [s:u_st](le m (s u_time))).

Unfold u_pr.
Unfold u_co.
Intros.
Simpl.
Split.
Intros.
Rewrite -> (u_assign_changed s u_time (f (s u_time))).
Rewrite <- H.
Apply f_is_ascending.
Split.
Intros.
Rewrite -> (u_assign_changed s u_time (g (s u_time))).
Rewrite <- H.
Apply g_is_ascending.
Auto.
Qed.




Lemma u_pr_ne : (u_prog_not_empty u_pr).

Unfold u_prog_not_empty.
Exists  update_f.
Unfold u_pr.
Simpl.
Left.
Reflexivity.
Qed.


Definition variant := [s:u_st](s u_time).


Lemma CMT_S7 : (u_stable u_pr [s:u_st](le (variant s) cmt)).

Unfold variant.
Generalize (CMT_S3 cmt).
Intros.
Apply (u_stable_subst u_pr) with p:=[s:u_st]
                                     (common cmt)->(le (s u_time) cmt).
Unfold u_everywhere u_iff.
Intros.
Split.
Intros.
Apply (H0 cmt_exists).
Intros.
Auto.
Auto.
Qed.


Lemma CMT_S8 : (m:nat)
  (u_unless u_pr
     (u_and [s:u_st](le (s u_time) cmt) [s:u_st]m=(s u_time))
     (u_or
       (u_and [s:u_st](le (s u_time) cmt) [s:u_st](lt m (s u_time)))
       [s:u_st](s u_time)=cmt)).

Intros.
Generalize (u_co_stable_conjunction u_pr [s:u_st]m=(s u_time)
             [s:u_st](le m (s u_time)) [s:u_st](le (variant s) cmt)
             CMT_S7 (CMT_S6 m)).
Intros.
Unfold u_unless.
Apply u_co_implication with p':=(u_and [s:u_st]m=(s u_time)
                                  [s:u_st](le (variant s) cmt)) q:=(
u_and [s:u_st](le m (s u_time)) [s:u_st](le (variant s) cmt)).
Unfold u_everywhere u_imp u_or u_and.
Intros.
Unfold variant.
Elim H0.
Intros.
Elim H1.
Intros.
Split.
Auto.
Auto.
Unfold u_everywhere u_imp u_and u_or.
Intros.
Elim H0.
Intros.
Clear  H H0.
Unfold variant in H2.
Elim (eq_nat_dec m (s u_time)).
Intros.
Left.
Split.
Auto.
Auto.
Intros.
Right.
Left.
Split.
Auto.
Elim (le_lt_or_eq m (s u_time)).
Intros.
Auto.
Intros.
Elim y.
Auto.
Auto.
Auto.
Qed.


Lemma CMT_S9 : (m:nat) (u_transient u_pr 
  [s:u_st](not (common (s u_time))) /\ m=(s u_time)).

Intros.
Unfold u_transient.
Elim (eq_nat_dec m (f m)).
Intros.
Exists  update_g.
Split.
Simpl.
Right.
Left.
Auto.
Intros.
Elim H.
Intros.
Clear  H.
Simpl.
Unfold not.
Intros.
Elim H.
Intros.
Clear  H.
Rewrite (u_assign_changed s u_time (g (s u_time))) in H3.
Rewrite <- H1 in H0.
Rewrite <- H1 in H3.
Unfold common in H0.
Elim H0.
Split.
Auto.
Auto.
Intros.
Exists  update_f.
Split.
Simpl.
Left.
Auto.
Intros.
Simpl.
Rewrite -> (u_assign_changed s u_time (f (s u_time))).
Elim H.
Intros.
Clear  H.
Rewrite <- H1.
Unfold not.
Intros.
Elim H.
Intros.
Apply y.
Auto.
Qed.


Lemma CMT_S10 : (m:nat) (u_transient u_pr 
  [s:u_st]((lt (s u_time) cmt) /\ m=(s u_time))).

Intros.
Apply u_transient_implication with p:=[s:u_st]
                                       ~(common (s u_time))
                                       /\m=(s u_time).
Apply CMT_S9.
Unfold u_everywhere u_imp.
Intros.
Elim H.
Clear  H.
Intros.
Split.
Unfold not.
Intros.
Generalize (cmt_smallest (s u_time) H1).
Intros.
Generalize (le_lt_trans cmt (s u_time) cmt H2 H).
Intros.
Apply (lt_n_n cmt H3).
Auto.
Qed.


Lemma CMT_L12 : (m:nat)(u_transient u_pr
   (u_and (u_and [s:u_st](le (s u_time) cmt) [s:u_st]m=(s u_time))
     (u_not
       (u_or
         (u_and [s:u_st](le (s u_time) cmt) [s:u_st](lt m (s u_time)))
         [s:u_st](s u_time)=cmt)))).

Intros.
Apply u_transient_implication with p:=[s:u_st]
                                       (lt (s u_time) cmt)
                                       /\m=(s u_time).
Apply CMT_S10.
Unfold u_everywhere u_imp u_and u_or u_not.
Intros.
Elim H.
Clear  H.
Intros.
Elim H.
Clear  H.
Intros.
Split.
Elim (eq_nat_dec (s u_time) cmt).
Intros.
Elim H0.
Right.
Auto.
Intros.
Elim (le_lt_or_eq (s u_time) cmt H).
Intros.
Auto.
Intros.
Elim y.
Auto.
Auto.
Qed.


Lemma CMT_L13 : (m:nat)
  (u_leads_to u_pr
    (u_and [s:u_st](le (s u_time) cmt) [s:u_st]m=(variant s))
    (u_or
      (u_and [s:u_st](le (s u_time) cmt) [s:u_st](lt m (variant s)))
      [s:u_st](s u_time)=cmt)).

Intros.
Unfold variant.
Apply u_leads_to_basis.
Unfold u_ensures.
Split.
Apply CMT_S8.
Apply CMT_L12.
Qed.


Lemma CMT_L14 : (u_leads_to u_pr 
  ([s:u_st](le (s u_time) cmt)) ([s:u_st](s u_time)=cmt)).

Apply (u_leads_to_rev_nat_induction u_pr u_pr_ne variant cmt).
Apply u_imp_p_p.
Intros.
Apply CMT_L13.
Qed.


Theorem CMT_L15 : (u_leads_to u_pr 
  ([s:u_st](s u_time)=O) ([s:u_st](s u_time)=cmt)).

Apply (u_leads_to_implication_l u_pr u_pr_ne [s:u_st](s u_time)=O
        [s:u_st](le (s u_time) cmt) [s:u_st](s u_time)=cmt).
Unfold u_everywhere u_imp.
Intros.
Rewrite -> H.
Apply le_O_n.
Apply CMT_L14.
Qed.




