(*C*
      Embedding UNITY into the Calculus of Inductive Constructions
                                     Using the COQ Proof Assistent
      Author:  Mark-Oliver Stehr
               University of Hamburg
               Department of Computer Science
               Theoretical Foundations of Computer Science Group 
               stehr@informatik.uni-hamburg.de
      Date:    September 1998
*C*)





Theorem u_var_index_thm : (P:u_var->Prop)
  ((i:nat)(lt i u_vars)->(P (u_ith_var i)))->(v:u_var)(P v).

Intros.
Generalize (u_var_finite v).
Intro.
Elim H0.
Intros i H1.
Elim H1.
Intros.
Rewrite <- H3.
Apply H.
Assumption.
Qed.


Fixpoint u_equalize_n_vars [n:nat] : u_st->u_st->u_st :=
  [s,s':u_st] Case n of 
    (* O *) s 
    (* (S p) *) [p:nat] (u_assign (u_equalize_n_vars p s s') 
                                  (u_ith_var p) (s' (u_ith_var p)))
  end.


Lemma u_equalize_1 : (s,s':u_st)(n:nat)(i:nat)(lt i n)->
((u_equalize_n_vars n s s') (u_ith_var i))=(s' (u_ith_var i)).

Intros s s'.
Induction n.
Intros.
Apply False_ind.
Apply (lt_n_O i H).
Intros.
Simpl.
Generalize (lt_n_Sm_le i n0 H0).
Intro.
Generalize (le_lt_or_eq i n0 H1).
Intro.
Elim H2.
Intro.
Elim (u_var_eq_dec (u_ith_var n0) (u_ith_var i)).
Intros H4.
Rewrite -> H4.
Rewrite -> u_assign_changed.
Reflexivity.
Intro.
Rewrite -> u_assign_indep.
Apply H.
Assumption.
Assumption.
Intro.
Rewrite <- H3.
Rewrite -> u_assign_changed.
Reflexivity.
Qed.


Theorem u_equalize_works : (s,s':u_st) (u_equalize_n_vars u_vars s s')=s'.

Intros.
Apply u_st_extensionality.
Intro.
Apply (u_var_index_thm
        [v:u_var](u_equalize_n_vars u_vars s s' v)=(s' v)).
Intros i H.
Apply u_equalize_1.
Assumption.
Qed.


Definition u_independent := [p:u_prop][v:u_var]
  (s:u_st)(x,y:(u_var_type v)) (p (u_assign s v x)) -> (p (u_assign s v y)).


Theorem  u_independent_imp : (p:u_prop)(v:u_var)(u_independent p v)->
  (s:u_st)(p s)->(x:(u_var_type v))(p (u_assign s v x)).

Intros.
Unfold u_independent in H.
Apply (H s (s v) x).
Rewrite -> u_assign_unchanged.
Assumption.
Qed.


Lemma u_independent_and_equalize : (p:u_prop)
  ((v:u_var) (u_independent p v))->
  (s,s':u_st)(i:nat)(le i u_vars)->
  (p s)->(p (u_equalize_n_vars i s s')).

Intros p H s s'.
Induction i.
Intros.
Simpl.
Assumption.
Intros.
Simpl.
Generalize (H (u_ith_var n)).
Intro.
Generalize (u_independent_imp p (u_ith_var n) H3).
Intro.
Apply H4.
Apply H0.
Apply le_trans with (S n).
Apply le_n_Sn.
Assumption.
Assumption.
Qed.


Theorem u_fully_independent : (p:u_prop)
  ((v:u_var)(u_independent p v))->(s,s':u_st)(p s)->(p s').

Intros.
Rewrite <- (u_equalize_works s s').
Apply u_independent_and_equalize.
Assumption.
Apply le_n.
Assumption.
Qed.


Definition u_may_depend_on := [p:u_prop][v:u_var]
  (m:(u_var_type v))(s,s':u_st) (p (u_assign s v m))->(p (u_assign s' v m)).


Lemma u_make_independent : (p:u_prop)(v:u_var)(m:(u_var_type v))
  (u_independent [s:u_st](p (u_assign s v m)) v).

Intros.
Unfold u_independent.
Intros s x y.
Rewrite -> u_assign_twice.
Rewrite -> u_assign_twice.
Intro.
Assumption.
Qed.


Theorem u_dependent : (x:u_var)(p:u_prop)
  ((v:u_var)~(v=x)->(u_independent p v)) -> (u_may_depend_on p x).

Intros.
Unfold u_may_depend_on.
Intros.
Generalize (u_fully_independent [s:u_st](p (u_assign s x m))).
Intro.
Apply H1 with s:=s.
Intro.
Elim (u_var_eq_dec v x).
Intros H2.
Rewrite -> H2.
Apply u_make_independent.
Intro.
Generalize (H v y).
Intro.
Unfold u_independent.
Unfold u_independent in H2.
Intros s'' x0 y0.
Rewrite -> u_assign_commutes.
Pattern (u_assign (u_assign s'' v y0) x m) .
Rewrite -> u_assign_commutes.
Intro.
Apply (H2 (u_assign s'' x m) x0 y0).
Assumption.
Assumption.
Assumption.
Assumption.
Qed.


Definition u_with := [p:u_prop][v:u_var][x:(u_var_type v)]
  [s:u_st](p (u_assign s v x)).  


Theorem u_indep_with : (p:u_prop)(v:u_var)(m:(u_var_type v))
  (u_independent (u_with p v m) v).

Intros.
Unfold u_with.
Apply u_make_independent.
Qed.


Theorem u_indep_imp_indep_with : (p:u_prop)(v,v':u_var)(m:(u_var_type v))
  (u_independent p v')->(u_independent (u_with p v m) v').

Intros.
Unfold u_with.
Unfold u_independent.
Unfold u_independent in H.
Intros s.
Elim (u_var_eq_dec v v').
Intros H1.
Rewrite <- H1.
Intros x y.
Rewrite -> u_assign_twice.
Rewrite -> u_assign_twice.
Intro.
Assumption.
Intros H1 x y.
Rewrite -> u_assign_commutes.
Pattern (u_assign (u_assign s v' y) v m) .
Rewrite -> u_assign_commutes.
Intro.
Apply (H (u_assign s v m) x y).
Assumption.
Unfold not.
Intro.
Rewrite H0 in H1.
Apply H1.
Reflexivity.
Unfold not.
Intro.
Rewrite H0 in H1.
Apply H1.
Reflexivity.
Qed.




