(*C*
      Embedding UNITY into the Calculus of Inductive Constructions
                                     Using the COQ Proof Assistent
      Author:  Mark-Oliver Stehr
               University of Hamburg
               Department of Computer Science
               Theoretical Foundations of Computer Science Group 
               stehr@informatik.uni-hamburg.de
      Date:    September 1998
*C*)




Definition u_transient := [pr:u_prog][p:u_prop]
  (Ex [t:u_tr](In t pr)/\((s:u_st)(p s)->~(p ((u_effect t) s)))).


Theorem u_transient_false : (pr:u_prog)
  (u_prog_not_empty pr)->(u_transient pr u_false).

Intros.
Unfold u_transient.
Unfold u_prog_not_empty in H.
Elim H.
Intros t.
Intros.
Exists  t.
Split.
Assumption.
Intros.
Unfold u_false in H1.
Elim H1.
Qed.


Theorem u_transient_implication : (pr:u_prog)(p,q:u_prop)
  (u_transient pr p)->(u_everywhere (u_imp q p))->(u_transient pr q).

Unfold u_transient.
Intros.
Elim H.
Intros t.
Intros.
Exists  t.
Split.
Elim H1.
Intros.
Assumption.
Elim H1.
Intros.
Clear  H1.
Unfold not.
Intros.
Unfold not in H3.
Apply (H3 s).
Apply (H0 s).
Assumption.
Apply (H0 (u_effect t s)).
Assumption.
Qed.


Theorem u_transient_strengthen : (pr:u_prog)(p,q:u_prop)
  (u_transient pr p)->(u_transient pr (u_and p q)).

Unfold u_transient.
Intros.
Elim H.
Intros t.
Intros.
Exists  t.
Unfold u_and.
Intros.
Split.
Elim H0.
Intros.
Assumption.
Intros.
Elim H1.
Intros.
Unfold not.
Intros.
Elim H0.
Intros.
Generalize (H6 s H2).
Elim H4.
Intros.
Apply (H9 H7).
Qed.


Theorem u_stable_and_transient : (pr:u_prog)(p:u_prop)
  (u_stable pr p) -> (u_transient pr p) -> (u_everywhere (u_not p)).

Unfold u_stable.
Unfold u_transient.
Unfold u_everywhere u_not.
Intros.
Elim H0.
Intros t.
Intros.
Clear  H0.
Elim H1.
Intros.
Clear  H1.
Generalize (H2 s).
Intros.
Clear  H2.
Unfold not.
Intros.
Generalize (H1 H2).
Intros.
Clear  H1.
Generalize (u_nc_co pr p p H).
Intros.
Generalize (H1 t s H0 H2).
Intros.
Apply (H3 H4).
Qed.


Theorem u_transient_subst : (pr:u_prog)(p,q:u_prop)
  (u_everywhere (u_iff p q)) ->
  (u_transient pr p) -> (u_transient pr q).

Unfold u_everywhere u_iff u_transient.
Intros.
Elim H0.
Intros t.
Intros.
Exists  t.
Split.
Elim H1.
Intros.
Assumption.
Intros.
Elim H1.
Intros.
Generalize (H4 s).
Intros.
Generalize (H s).
Intros.
Elim H6.
Intros.
Generalize (H8 H2).
Intros.
Generalize (H5 H9).
Intros.
Unfold not.
Intros.
Generalize (H (u_effect t s)).
Intros.
Elim H12.
Intros.
Generalize (H14 H11).
Intros.
Apply (H10 H15).
Qed.


Theorem u_rel_transient : (pr:u_prog)(u_prog_not_empty pr)->(p,r:u_prop)
  (u_stable pr r)->
  (Ex [t:u_tr](In t pr)/\
    ((s:u_st)(r s)->(p s)->(r ((u_effect t) s))->~(p ((u_effect t) s))))->
  (u_transient pr (u_and r p)).

Intros.
Unfold u_transient.
Elim H1.
Clear H1.
Intro t.
Intros.
Split with t.
Unfold u_stable in H0.
Generalize (u_nc_co pr ? ? H0).
Clear H0.
Intros.
Generalize (H0 t).
Clear H0.
Intros.
Elim H1.
Clear H1.
Intros.
Split.
Trivial.
Intros.
Unfold u_and.
Unfold not.
Intros.
Elim (H2 s).
Elim H3.
Auto.
Elim H3.
Auto.
Elim H4.
Auto.
Elim H4.
Auto.
Qed.


Definition u_ensures := [pr:u_prog][p,q:u_prop]
  (u_unless pr p q) /\ (u_transient pr (u_and p (u_not q))).


Theorem u_ensures_true : (pr:u_prog)(u_prog_not_empty pr)->
  (p:u_prop)(u_ensures pr p u_true).

Intros.
Unfold u_ensures.
Split.
Apply u_unless_true.
Apply u_transient_implication with p:=u_false.
Apply u_transient_false.
Trivial.
Unfold u_everywhere u_imp u_and u_not u_false.
Tauto.
Qed.


Theorem u_ensures_subst_r : (pr:u_prog)(p,q,q':u_prop)
  (u_everywhere (u_iff q q'))->(u_ensures pr p q)->(u_ensures pr p q').

Unfold u_ensures.
Intros.
Elim H0.
Intros.
Split.
Apply u_unless_subst_r with q:=q.
Assumption.
Assumption.
Apply (u_transient_subst pr) with p:=(u_and p (u_not q)).
Unfold u_everywhere u_and u_not u_iff.
Unfold u_everywhere u_iff in H.
Intros.
Elim (H s).
Intros.
Split.
Intros.
Elim H5.
Intros.
Split.
Assumption.
Unfold not.
Unfold not in H7.
Intros.
Apply H7.
Apply H4.
Assumption.
Intros.
Elim H5.
Intros.
Split.
Assumption.
Unfold not.
Unfold not in H7.
Intros.
Apply H7.
Apply H3.
Assumption.
Assumption.
Qed.

Theorem u_ensures_subst_l : (pr:u_prog)(p,p',q:u_prop)
  (u_everywhere (u_iff p p'))->(u_ensures pr p q)->(u_ensures pr p' q).

Unfold u_ensures.
Intros.
Elim H0.
Intros.
Split.
Apply u_unless_subst_l with p:=p.
Assumption.
Assumption.
Apply (u_transient_subst pr) with p:=(u_and p (u_not q)).
Unfold u_everywhere u_and u_not u_iff.
Unfold u_everywhere u_iff in H.
Intros.
Elim (H s).
Intros.
Split.
Intros.
Elim H5.
Intros.
Split.
Apply H3.
Assumption.
Assumption.
Intros.
Elim H5.
Intros.
Split.
Apply H4.
Assumption.
Assumption.
Assumption.
Qed.


Theorem u_ensures_refl : (pr:u_prog)(u_prog_not_empty pr)->
  (p:u_prop)(u_ensures pr p p).

Unfold u_ensures.
Intros.
Split.
Apply u_unless_refl.
Apply u_transient_subst with p:=u_false.
Unfold u_everywhere u_iff u_false u_and u_not.
Intros.
Split.
Intros.
Elim H0.
Intros.
Elim H0.
Intros.
Apply (H2 H1).
Apply u_transient_false.
Assumption.
Qed.


Theorem u_ensures_imp : (pr:u_prog)(u_prog_not_empty pr)->
  (p,q:u_prop)(u_everywhere (u_imp p q)) -> (u_ensures pr p q).

Intros.
Unfold u_ensures.
Split.
Apply u_unless_imp.
Assumption.
Apply u_transient_subst with p:=u_false.
Unfold u_everywhere u_iff u_false u_and u_not.
Intros.
Split.
Intro.
Elim H1.
Intros.
Elim H1.
Intros.
Unfold u_everywhere u_imp in H0.
Generalize (H0 s H2).
Intros.
Apply (H3 H4).
Apply u_transient_false.
Assumption.
Qed.


Theorem u_ensures_false : (pr:u_prog)(u_prog_not_empty pr)->
  (p:u_prop)(u_ensures pr p u_false)->(u_everywhere (u_not p)).

Intros.
Unfold u_ensures in H0.
Elim H0.
Intros.
Unfold u_transient in H2.
Elim H2.
Intros t.
Intros.
Elim H3.
Intros.
Unfold u_everywhere u_not.
Unfold not.
Intros.
Generalize (H5 s).
Unfold u_and u_not u_false.
Intros.
Unfold u_unless in H1.
Cut (u_co pr p p).
Intros.
Generalize (u_nc_co pr p p H8).
Intros.
Generalize (H9 t s).
Intros.
Generalize (H10 H4).
Intros.
Unfold not in H7.
Apply H7.
Split.
Assumption.
Auto.
Split.
Apply (H11 H6).
Auto.
Apply u_co_implication with p':=(u_and p (u_not u_false)) q:=(u_or p
                                                             u_false).
Unfold u_everywhere u_imp u_and u_not u_false.
Intros.
Split.
Assumption.
Unfold not.
Auto.
Unfold u_everywhere u_imp u_or u_false.
Intros.
Elim H8.
Auto.
Intro.
Elim H9.
Assumption.
Qed.


Theorem u_ensures_implication_r : (pr:u_prog)(p,q,r:u_prop)
  (u_everywhere (u_imp q r)) ->
  (u_ensures pr p q) -> (u_ensures pr p r).

Unfold u_ensures.
Intros.
Elim H0.
Intros.
Split.
Apply u_unless_implication_r with q:=q.
Assumption.
Assumption.
Clear  H0.
Apply u_transient_implication with p:=(u_and p (u_not q)).
Assumption.
Unfold u_everywhere u_imp u_and u_not.
Intros.
Elim H0.
Intros.
Split.
Assumption.
Unfold not.
Intros.
Apply H4.
Apply (H s).
Assumption.
Qed.


Theorem u_ensures_conjunction : (pr:u_prog)(p,p',q,q':u_prop)
  (u_unless pr p q) -> (u_ensures pr p' q') -> 
  (u_ensures pr (u_and p p') 
     (u_or3 (u_and p q') (u_and p' q) (u_and q q'))).

Unfold u_ensures.
Intros.
Split.
Apply u_unless_conjunction.
Assumption.
Elim H0.
Intros.
Assumption.
Elim H0.
Intros.
Clear  H0.
Apply u_transient_implication with p:=(u_and p' (u_not q')).
Assumption.
Unfold u_everywhere u_imp u_and u_not.
Intros.
Unfold u_or3 in H0.
Elim H0.
Intros.
Split.
Elim H3.
Intros.
Assumption.
Unfold not.
Intros.
Apply H4.
Left.
Split.
Elim H3.
Intros.
Assumption.
Assumption.
Qed.


Theorem u_rel_ensures : (pr:u_prog)(u_prog_not_empty pr)->(p,q,r:u_prop)
  (u_stable pr r)->
  (u_unless pr (u_and r p) (u_imp r q))->
  (u_transient pr (u_and r (u_and p (u_not q))))->
  (u_ensures pr (u_and r p) (u_and r q)).

Unfold u_ensures.
Intros.
Split.
Apply u_rel_unless.
Trivial.
Unfold u_unless in H1.
Apply u_co_implication with 3:=H1.
Unfold u_everywhere u_imp u_and u_and3 u_not.
Clear H0 H1 H2.
Tauto.
Unfold u_everywhere u_imp u_and u_or.
Clear H0 H1 H2 H.
Tauto.
Apply u_transient_subst with 2:=H2.
Unfold u_everywhere u_iff u_and u_not.
Clear H H0 H1 H2.
Tauto.
Qed.


Inductive u_leads_to [pr:u_prog] : u_prop->u_prop->Prop :=
  u_leads_to_basis : (p,q:u_prop)
  (u_ensures pr p q) -> (u_leads_to pr p q)
| u_leads_to_transitivity : (p,q,r:u_prop)
  (u_leads_to pr p q) -> (u_leads_to pr q r) -> (u_leads_to pr p r)
| u_leads_to_disjunction : (T:Set)(P:T->u_prop)(p,q:u_prop)
  ((i:T)(u_leads_to pr (P i) q)) -> (u_everywhere (u_iff p (u_ex ? P))) ->
  (u_leads_to pr p q).





Theorem u_leads_to_true : (pr:u_prog)(u_prog_not_empty pr)->
  (p:u_prop)(u_leads_to pr p u_true).

Intros.
Apply u_leads_to_basis.
Apply u_ensures_true.
Trivial.
Qed.


Theorem u_leads_to_disjunction_2 : 
  (pr:u_prog)(T:Set)(P:T->u_prop)(q:u_prop)
  ((i:T)(u_leads_to pr (P i) q)) -> (u_leads_to pr (u_ex ? P) q).

Intros.
Apply (u_leads_to_disjunction pr T P (u_ex T P) q).
Assumption.
Apply u_iff_refl.
Qed.


Theorem u_leads_to_subst_r : (pr:u_prog)(p,q,q':u_prop)
  (u_everywhere (u_iff q q'))->(u_leads_to pr p q)->(u_leads_to pr p q').

Intros pr.
Cut (p,q,q':u_prop)
     (u_leads_to pr p q)
      ->(u_everywhere (u_iff q q'))->(u_leads_to pr p q').
Intros.
Generalize (H p q q' H1 H0).
Auto.
Induction 1.
Intros.
Constructor 1.
Apply u_ensures_subst_r with p:=p0 q:=q0 q':=q'.
Exact H1.
Exact H0.
Intros.
Generalize (H3 H4).
Intros.
Constructor 2 with q:=q0.
Exact H0.
Exact H5.
Intros.
Constructor 3 with P:=P.
Intros.
Generalize (H1 i H3).
Intros.
Exact H4.
Exact H2.
Qed.


Theorem u_leads_to_subst_l : (pr:u_prog)(p,p',q:u_prop)
  (u_everywhere (u_iff p p'))->(u_leads_to pr p q)->(u_leads_to pr p' q).

Intros pr.
Cut (p,q,p':u_prop)
     (u_leads_to pr p q)
      ->(u_everywhere (u_iff p p'))->(u_leads_to pr p' q).
Intros.
Generalize (H p q p' H1 H0).
Intros.
Exact H2.
Induction 1.
Intros.
Constructor 1.
Apply u_ensures_subst_l with p:=p0.
Exact H1.
Exact H0.
Intros.
Constructor 2 with q:=q0.
Generalize (H1 H4).
Intros.
Exact H5.
Exact H2.
Intros.
Constructor 3 with P:=P.
Exact H0.
Apply u_iff_trans with q:=p0.
Apply u_iff_sym.
Exact H3.
Exact H2.
Qed.


Inductive u_lt_pre [pr:u_prog;r:u_prop] : u_prop->Prop :=
  u_lt_pre_basis : (p:u_prop)
  (u_ensures pr p r) -> (u_lt_pre pr r p)
| u_lt_pre_transitivity : (p,q:u_prop)
  (u_ensures pr p q) -> (u_lt_pre pr r q) -> (u_lt_pre pr r p)
| u_lt_pre_disjunction : (T:Set)(P:T->u_prop)
  ((i:T)(u_lt_pre pr r (P i))) -> 
  (p:u_prop)(u_everywhere (u_iff p (u_ex ? P))) -> 
  (u_lt_pre pr r p).


Theorem u_lt_pre_imp_leads_to : (pr:u_prog)(p,q:u_prop)
  (u_lt_pre pr q p) -> (u_leads_to pr p q).

Intros.
Elim H.
Intros.
Constructor 1.
Assumption.
Intros.
Cut (u_leads_to pr p0 q0).
Intros.
Apply u_leads_to_transitivity with q:=q0.
Assumption.
Assumption.
Constructor 1.
Assumption.
Intros.
Constructor 3 with P:=P.
Assumption.
Assumption.
Qed.

Theorem u_leads_to_imp_precond : (pr:u_prog)(p,q:u_prop)
  (u_leads_to pr p q) -> (u_lt_pre pr q p).

Intros.
Elim H.
Intros.
Constructor 1.
Assumption.
Intros.
Elim H1.
Intros.
Constructor 2 with q:=q0.
Assumption.
Assumption.
Intros.
Constructor 2 with q:=q1.
Assumption.
Assumption.
Intros.
Constructor 3 with P:=P.
Assumption.
Assumption.
Intros.
Constructor 3 with P:=P.
Assumption.
Assumption.
Qed.


Theorem u_leads_to_refl : (pr:u_prog)(u_prog_not_empty pr)->
  (p:u_prop)(u_leads_to pr p p).

Intros.
Apply u_leads_to_basis.
Apply u_ensures_refl.
Assumption.
Qed.


Theorem u_leads_to_imp : (pr:u_prog)(u_prog_not_empty pr)->
  (p,q:u_prop)(u_everywhere (u_imp p q)) -> (u_leads_to pr p q).

Intros.
Apply u_leads_to_basis.
Apply u_ensures_imp.
Assumption.
Assumption.
Qed.


Theorem u_leads_to_false : (pr:u_prog)(u_prog_not_empty pr)->
  (p:u_prop)(u_leads_to pr u_false p).

Intros.
Apply u_leads_to_imp.
Assumption.
Apply u_imp_false.
Qed.


Theorem u_leads_to_implication_l : (pr:u_prog)(u_prog_not_empty pr)->
  (p,q,r:u_prop)(u_everywhere (u_imp p q)) ->
  (u_leads_to pr q r) -> (u_leads_to pr p r).

Intros.
Apply u_leads_to_transitivity with q:=q.
Apply u_leads_to_imp.
Assumption.
Assumption.
Assumption.
Qed.

Theorem u_leads_to_implication_r : (pr:u_prog)(u_prog_not_empty pr)->
  (p,q,r:u_prop)(u_everywhere (u_imp q r)) ->
  (u_leads_to pr p q) -> (u_leads_to pr p r).

Intros.
Apply u_leads_to_transitivity with q:=q.
Assumption.
Apply u_leads_to_imp.
Assumption.
Assumption.
Qed.


Theorem u_leads_to_implication : (pr:u_prog)(u_prog_not_empty pr)->
  (p,p',q,q':u_prop)
  (u_everywhere (u_imp p p'))->
  (u_everywhere (u_imp q q'))->
  (u_leads_to pr p' q)-> (u_leads_to pr p q').

Intros.
Apply u_leads_to_transitivity with q:=p'.
Apply u_leads_to_imp.
Assumption.
Assumption.
Apply u_leads_to_transitivity with q:=q.
Assumption.
Apply u_leads_to_imp.
Assumption.
Assumption.
Qed.


Theorem u_leads_to_weakening : (pr:u_prog)(u_prog_not_empty pr)->
  (p,q,r:u_prop)(u_leads_to pr p q)->(u_leads_to pr p (u_or q r)).

Intros.
Apply u_leads_to_implication_r with q:=q.
Assumption.
Apply u_imp_or_l.
Assumption.
Qed.

Theorem u_leads_to_strengthening : (pr:u_prog)(u_prog_not_empty pr)->
  (p,q,r:u_prop)(u_leads_to pr p q)->(u_leads_to pr (u_and p r) q).

Intros.
Apply u_leads_to_implication_l with q:=p.
Assumption.
Apply u_imp_and_l.
Assumption.
Qed.


Theorem u_leads_to_binary_disjunction : 
  (pr:u_prog)(u_prog_not_empty pr)->(p,p',q:u_prop)
  (u_leads_to pr p q)->(u_leads_to pr p' q)-> 
  (u_leads_to pr (u_or p p') q).

Intros.
Generalize (u_leads_to_disjunction_2 pr bool
             [b:bool]Cases b of
                       true => p
                     | false => p'
                     end q).
Intros.
Cut (u_leads_to pr (u_ex bool [b:bool]Case b of p
                                                p'
                                                end) q).
Intros.
Apply u_leads_to_implication_l
 with q:=(u_ex bool [b:bool]Case b of p
                                      p'
                                      end).
Trivial.
Unfold u_everywhere u_or u_ex.
Intros.
Unfold u_imp.
Intros.
Elim H4.
Intros.
Split with true.
Trivial.
Intros.
Split with false.
Trivial.
Trivial.
Apply H2.
Induction i.
Trivial.
Trivial.
Qed.

Theorem u_leads_to_ternary_disjunction : 
  (pr:u_prog)(u_prog_not_empty pr)->(p,p',p'',q:u_prop)
  (u_leads_to pr p q)->(u_leads_to pr p' q)->(u_leads_to pr p'' q)-> 
  (u_leads_to pr (u_or3 p p' p'') q).

Intros.
Specialize (u_leads_to_binary_disjunction pr H ? ? ? H0 H1).
Intros.
Specialize (u_leads_to_binary_disjunction pr H ? ? ? H3 H2).
Intros.
Apply u_leads_to_implication_l with q:=(u_or (u_or p p') p'').
Trivial.
Trivial.
Unfold u_everywhere u_imp u_or3 u_or.
Tauto.
Trivial.
Qed.


Theorem u_leads_to_big_disjunction : (pr:u_prog)(u_prog_not_empty pr)->
  (T:Set)(P,Q:T->u_prop)
  ((i:T)(u_leads_to pr (P i) (Q i)))->
  (u_leads_to pr (u_ex ? P) (u_ex ? Q)).

Intros.
Cut (m:T)(u_leads_to pr (Q m) (u_ex ? Q)).
Intros.
Cut (i:T)(u_leads_to pr (P i) (u_ex T Q)).
Intros.
Apply u_leads_to_disjunction with P:=P.
Assumption.
Apply u_iff_refl.
Intro.
Generalize (H0 i).
Intros.
Generalize (H1 i).
Intros.
Apply u_leads_to_transitivity with 1:=H2 2:=H3.
Intros.
Apply u_leads_to_imp.
Assumption.
Unfold u_everywhere u_imp u_ex.
Intros.
Exists  m.
Assumption.
Qed.


Theorem u_leads_to_small_disjunction : 
  (pr:u_prog)(u_prog_not_empty pr) -> (p,q,p',q':u_prop)
  (u_leads_to pr p p') -> (u_leads_to pr q q')->
  (u_leads_to pr (u_or p q) (u_or p' q')).

Intros.
Generalize (u_leads_to_big_disjunction pr H bool
             [b:bool]Cases b of 
                         true  => p
                      |  false  => q
                     end
             [b:bool]Cases b of 
                         true  => p'
                      |  false  => q'
                     end).
Intros.
Apply u_leads_to_subst_r with q:=(u_ex bool
                                   [b:bool]<u_prop>Case b of p'
                                                             q'
                                                             end).
Unfold u_everywhere u_iff u_ex u_or.
Intros.
Split.
Intros.
Elim H3.
Induction x.
Intros.
Left.
Assumption.
Intros.
Right.
Assumption.
Intros.
Elim H3.
Intros.
Exists  true.
Assumption.
Intros.
Exists  false.
Assumption.
Apply u_leads_to_subst_l with p:=(u_ex bool
                                   [b:bool]<u_prop>Case b of p
                                                             q
                                                             end).
Unfold u_everywhere u_iff u_ex u_or.
Intros.
Split.
Intros.
Elim H3.
Induction x.
Intros.
Left.
Assumption.
Intros.
Right.
Assumption.
Intros.
Elim H3.
Intros.
Exists  true.
Assumption.
Intros.
Exists  false.
Assumption.
Apply H2.
Induction i.
Assumption.
Assumption.
Qed.


Theorem u_leads_to_impossible : (pr:u_prog)(u_prog_not_empty pr)->
  (p:u_prop)(u_leads_to pr p u_false)->(u_everywhere (u_not p)).

Intros.
Generalize (u_leads_to_imp_precond pr p u_false H0).
Intros.
Clear  H0.
Elim H1.
Intros.
Apply u_ensures_false with pr:=pr.
Assumption.
Assumption.
Intros.
Apply u_ensures_false with pr:=pr.
Assumption.
Apply u_ensures_subst_r with q:=q.
Unfold u_everywhere u_iff u_false.
Unfold u_everywhere u_not in H3.
Intros.
Generalize (H3 s).
Intros.
Split.
Unfold not in H4.
Assumption.
Intro.
Elim H5.
Assumption.
Intros.
Unfold u_everywhere u_not u_ex.
Unfold u_everywhere u_not in H2.
Intros.
Unfold not.
Intros.
Generalize (H3 s).
Intros.
Elim H5.
Intros.
Generalize (H6 H4).
Intros.
Elim H8.
Intros i.
Intros.
Generalize (H2 i s).
Intros.
Apply H10.
Assumption.
Qed.


Theorem u_leads_to_cancelation : (pr:u_prog)(u_prog_not_empty pr) ->
  (p,q,b,r:u_prop)(u_leads_to pr p (u_or q b))->(u_leads_to pr b r)->
  (u_leads_to pr p (u_or q r)).

Intros.
Cut (u_leads_to pr (u_or q b) (u_or q r)).
Intros.
Apply u_leads_to_transitivity with q:=(u_or q b).
Assumption.
Assumption.
Apply u_leads_to_small_disjunction.
Apply H.
Apply (u_leads_to_refl pr H).
Assumption.
Qed.



Theorem u_ensures_psp : (pr:u_prog)(u_prog_not_empty pr)->
  (p,q,r,b:u_prop)(u_ensures pr p q)->(u_unless pr r b)->
  (u_ensures pr (u_and p r) (u_or (u_and q r) b)).

Intros.
Generalize (u_ensures_conjunction pr r p b q).
Intros.
Generalize (H2 H1 H0).
Intros.
Clear H2.
Apply u_ensures_subst_l with p:=(u_and r p).
Unfold u_everywhere u_iff u_and.
Intros.
Split.
Intros.
Elim H3.
Intros.
Split.
Elim H2.
Auto.
Elim H2.
Auto.
Intros.
Elim H2.
Intros.
Auto.
Apply u_ensures_implication_r
 with q:=(u_or3 (u_and r q) (u_and p b) (u_and b q)).
Unfold u_everywhere u_imp u_and u_or u_or3.
Intros.
Elim H2.
Intros.
Elim H4.
Intros.
Auto.
Intros.
Elim H4.
Intros.
Elim H5.
Intros.
Auto.
Intros.
Elim H4.
Intros.
Elim H6.
Auto.
Intros.
Elim H6.
Intros.
Auto.
Trivial.
Qed.


Theorem u_leads_to_psp : (pr:u_prog)(u_prog_not_empty pr)->
  (p,q,r,b:u_prop)(u_leads_to pr p q)->(u_unless pr r b)->
  (u_leads_to pr (u_and p r) (u_or (u_and q r) b)).

Intros.
Elim H0.
Intros.
Constructor 1.
Generalize (u_ensures_conjunction pr r p0 b q0).
Intros.
Generalize (H3 H1 H2).
Intros.
Clear  H3.
Apply u_ensures_subst_l with p:=(u_and r p0).
Unfold u_everywhere u_iff u_and.
Intros.
Split.
Intros.
Elim H3.
Intros.
Split.
Assumption.
Assumption.
Intros.
Elim H3.
Intros.
Split.
Assumption.
Assumption.
Apply u_ensures_implication_r with q:=(u_or3 (u_and r q0) (u_and p0 b)
                                        (u_and b q0)).
Unfold u_everywhere u_imp u_and u_or u_or3.
Intros.
Elim H3.
Intros.
Elim H5.
Intros.
Left.
Split.
Assumption.
Assumption.
Intros.
Elim H5.
Intros.
Right.
Elim H6.
Intros.
Assumption.
Intros.
Right.
Elim H6.
Intros.
Assumption.
Assumption.
Intros.
Cut (u_leads_to pr (u_and p0 r) (u_or b (u_and q0 r))).
Intros.
Generalize (u_leads_to_cancelation pr H (u_and p0 r) b (u_and q0 r)
             (u_or (u_and r0 r) b) H6 H5).
Intros.
Apply u_leads_to_subst_r with q:=(u_or b (u_or (u_and r0 r) b)).
Unfold u_everywhere u_iff u_or u_and.
Intros.
Split.
Intros.
Elim H8.
Intros.
Right.
Assumption.
Intros.
Assumption.
Intros.
Elim H8.
Intros.
Right.
Left.
Assumption.
Intros.
Left.
Assumption.
Assumption.
Apply u_leads_to_subst_r with q:=(u_or (u_and q0 r) b).
Unfold u_everywhere u_iff u_and u_or.
Intros.
Split.
Intros.
Elim H6.
Intros.
Right.
Assumption.
Intros.
Left.
Assumption.
Intros.
Elim H6.
Intros.
Right.
Assumption.
Intros.
Left.
Assumption.
Assumption.
Intros.
Generalize (u_leads_to_disjunction_2 pr T [i:T](u_and (P i) r)
             (u_or (u_and q0 r) b) H3).
Intros.
Apply u_leads_to_implication_l with q:=(u_ex T [i:T](u_and (P i) r)).
Assumption.
Unfold u_everywhere u_imp u_and u_or.
Intros.
Generalize (H4 s).
Intros.
Elim H7.
Intros.
Elim H6.
Intros.
Generalize (H8 H10).
Intros.
Elim H12.
Intros i.
Intros.
Exists  i.
Split.
Assumption.
Assumption.
Assumption.
Qed.


Theorem u_leads_to_stable_conjunction : (pr:u_prog)(u_prog_not_empty pr)->
  (p,q,r:u_prop)(u_stable pr r)->(u_leads_to pr p q)->
  (u_leads_to pr (u_and p r) (u_and q r)).

Intros.
Generalize (u_leads_to_psp pr H p q r u_false).
Intros.
Apply u_leads_to_transitivity with q:=(u_or (u_and q r) u_false).
Apply H2.
Assumption.
Apply u_stable_imp_unless.
Assumption.
Apply u_leads_to_imp.
Assumption.
Unfold u_everywhere u_imp u_or u_and u_false.
Intros.
Elim H3.
Intros.
Auto.
Intro.
Elim H4.
Qed.


Theorem u_rel_leads_to : (pr:u_prog)(u_prog_not_empty pr)->(p,q,r:u_prop)
  (u_stable pr r)->
  (u_leads_to pr (u_and r p) (u_imp r q))->
  (u_leads_to pr (u_and r p) (u_and r q)).

Intros.
Generalize (u_leads_to_stable_conjunction pr H ? ? ? H0 H1).
Clear H0 H1.
Intros.
Apply u_leads_to_implication with 4:=H0.
Trivial.
Clear H0 H.
Unfold u_everywhere u_imp u_and.
Tauto.
Clear H H0.
Unfold u_everywhere u_imp u_and.
Tauto.
Qed.


Theorem u_ind_invariant_imp_leads_to : (pr:u_prog)(p,q,r:u_prop)
  (u_prog_not_empty pr)->(u_ind_invariant pr p r)->
  (u_leads_to pr p q)->(u_leads_to pr p (u_and q r)).

Intros pr p q r H I H0.
Unfold u_ind_invariant in I.
Elim I.
Clear  I.
Intros.
Generalize (u_leads_to_stable_conjunction pr H p q r H2).
Intros.
Apply (u_leads_to_implication_l pr H) with q:=(u_and p r).
Unfold u_everywhere u_imp u_and.
Intros.
Split.
Auto.
Unfold u_everywhere u_imp in H1.
Apply (H1 s H4).
Apply H3.
Auto.
Qed.


Inductive u_strong_leads_to [pr:u_prog] : u_prop->u_prop->Prop :=
  u_strong_leads_to_basis : (p,q:u_prop)
  (u_ensures pr p q) -> (u_strong_leads_to pr p q)
| u_strong_leads_to_transitivity : (p,q,r:u_prop)
  (u_ensures pr p q) -> (u_strong_leads_to pr q r) -> 
                        (u_strong_leads_to pr p r)
| u_strong_leads_to_disjunction : (T:Set)(P:T->u_prop)(p,q:u_prop)
  ((i:T)(u_strong_leads_to pr (P i) q)) -> 
  (u_everywhere (u_iff p (u_ex ? P))) ->
  (u_strong_leads_to pr p q).


Theorem u_strong_leads_to_imp_leads_to :
  (pr:u_prog)(u_prog_not_empty pr)->(p,q:u_prop)
  (u_strong_leads_to pr p q)->(u_leads_to pr p q).

Intros.
Elim H0.
Intros.
Apply u_leads_to_basis.
Trivial.
Intros.
Cut (u_leads_to pr p0 q0).
Intros.
Apply u_leads_to_transitivity with q:=q0.
Trivial.
Trivial.
Trivial.
Apply u_leads_to_basis.
Trivial.
Intros.
Apply u_leads_to_disjunction with P:=P.
Intros.
Trivial.
Trivial.
Qed.

Theorem u_leads_to_imp_strong_leads_to :
  (pr:u_prog)(u_prog_not_empty pr)->(p,q:u_prop)
  (u_leads_to pr p q)->(u_strong_leads_to pr p q).

Intro.
Intro.
Induction 1.
Intros.
Apply u_strong_leads_to_basis.
Trivial.
Intros.
Generalize H4.
Clear H4.
Elim H1.
Intros.
Apply u_strong_leads_to_transitivity with q:=q1.
Trivial.
Trivial.
Clear H1 H2 H3.
Clear H0.
Intros.
Specialize (H3 H4).
Intros.
Apply H1.
Trivial.
Intros.
Apply u_strong_leads_to_disjunction with P:=P.
Intros.
Apply H5.
Trivial.
Trivial.
Intros.
Apply u_strong_leads_to_disjunction with P:=P.
Intros.
Apply H2.
Trivial.
Qed.


Theorem u_leads_to_strong_induction : 
  (IH:u_prog->u_prop->u_prop->Prop)(pr:u_prog)
  (u_prog_not_empty pr)->
  ((p,q:u_prop)(u_ensures pr p q)->(IH pr p q))->
  ((p,q,r:u_prop)
   ((u_ensures pr p q)->(u_leads_to pr q r)->
    (IH pr p q)->(IH pr q r)->(IH pr p r)))->
  ((T:Set)(P:T->u_prop)(p,q:u_prop)
   ((i:T)(u_leads_to pr (P i) q))->((i:T)(IH pr (P i) q))->
   (u_everywhere (u_iff p (u_ex ? P))) -> (IH pr p q))->  
  ((p,q:u_prop)
   (u_leads_to pr p q)->(IH pr p q)).

Intros.
Cut (u_strong_leads_to pr p q).
Intros.
Clear H3.
Elim H4.
Trivial.
Intros.
Apply H1 with q:=q0.
Trivial.
Apply u_strong_leads_to_imp_leads_to.
Trivial.
Trivial.
Apply H0.
Trivial.
Trivial.
Intros.
Apply H2 with P:=P.
Intros.
Trivial.
Trivial.
Trivial.
Trivial.
Apply u_strong_leads_to_imp_leads_to.
Trivial.
Apply H3.
Trivial.
Trivial.
Apply u_leads_to_imp_strong_leads_to.
Trivial.
Trivial.
Qed.


Theorem u_leads_to_double_induction : 
  (IH:u_prog->u_prop->u_prop->u_prop->u_prop->Prop)(pr:u_prog)
  (u_prog_not_empty pr)->
  ((p,q,p',q':u_prop)
   (u_ensures pr p q)->(u_leads_to pr p' q')->(IH pr p q p' q'))->
  ((p,q,p',q':u_prop)
   (u_leads_to pr p q)->(u_ensures pr p' q')->(IH pr p q p' q'))->
  ((p,q,r,p',q',r':u_prop)
   ((u_ensures pr p q)->(u_leads_to pr q r)->
    (u_ensures pr p' q')->(u_leads_to pr q' r')->
    (IH pr p q p' q')->
    (IH pr p q q' r')->(IH pr q r p' q')->
    (IH pr q r q' r')->
    (IH pr p r q' r')->(IH pr q r p' r')->
    (IH pr p r p' r')))->
  ((T:Set)(P':T->u_prop)(p,q,p',q':u_prop)
    (u_leads_to pr p q)->
    ((i:T)(u_leads_to pr (P' i) q'))->
    ((i:T)(IH pr p q (P' i) q'))->
    (u_everywhere (u_iff p' (u_ex ? P')))->
    (IH pr p q p' q'))->
  ((T:Set)(P:T->u_prop)(p,q,p',q':u_prop)
    ((i:T)(u_leads_to pr (P i) q))->
    (u_leads_to pr p' q')->
    ((i:T)(IH pr (P i) q p' q'))->
    (u_everywhere (u_iff p (u_ex ? P)))->
    (IH pr p q p' q'))->  
  ((p,q,p',q':u_prop)
   (u_leads_to pr p q)->(u_leads_to pr p' q')->(IH pr p q p' q')).

Intros IH pr NE H0 H1 H2 H3 H4.
Cut (p,q:u_prop)
     (u_leads_to pr p q)
      ->(p',q':u_prop)(u_leads_to pr p' q')->(IH pr p q p' q').
Clear H0 H1 H2 H3 H4.
Intros.
Apply H.
Trivial.
Trivial.
Generalize (u_leads_to_strong_induction
             [pr:u_prog]
              [p,q:u_prop]
               (p',q':u_prop)(u_leads_to pr p' q')->(IH pr p q p' q')
             pr NE).
Intros.
Apply H.
Intros.
Apply H0.
Trivial.
Trivial.
Intros.
Clear H.
Apply H2 with 1:=H7 2:=H8 q':=p'0.
Apply u_ensures_refl.
Trivial.
Trivial.
Generalize (H9 p'0 p'0).
Intros.
Apply H.
Apply u_leads_to_refl.
Trivial.
Apply H9.
Trivial.
Apply H1.
Trivial.
Apply u_ensures_refl.
Trivial.
Apply H10.
Trivial.
Generalize H11.
Generalize p'0 q'0.
Clear H11.
Clear p'0 q'0.
Generalize (u_leads_to_strong_induction
             [pr:u_prog][p'0,q'0:u_prop](IH pr p0 r p'0 q'0) pr NE).
Intros.
Apply H.
Intros.
Apply H1.
Apply u_leads_to_transitivity with q:=q0.
Apply u_leads_to_basis.
Trivial.
Trivial.
Trivial.
Clear H.
Intros.
Clear H13.
Clear H9.
Generalize (H10 q1 r0 H12).
Intros.
Specialize (u_leads_to_basis pr ? ? H).
Intros.
Specialize (u_leads_to_transitivity pr p1 q1 r0 H13 H12).
Intros.
Generalize (H10 p1 r0 H15).
Intros.
Apply (H2 p0 q0 r p1 q1 r0).
Trivial.
Trivial.
Trivial.
Trivial.
Apply H1.
Apply u_leads_to_basis.
Trivial.
Trivial.
Apply H0.
Trivial.
Trivial.
Apply H10.
Apply u_leads_to_basis.
Trivial.
Apply H10.
Trivial.
Trivial.
Apply H10.
Trivial.
Clear H.
Intros.
Apply H3 with P':=P.
Apply u_leads_to_transitivity with q:=q0.
Apply u_leads_to_basis.
Trivial.
Trivial.
Intros.
Apply H.
Trivial.
Trivial.
Trivial.
Apply H10.
Trivial.
Intros.
Apply H4 with P:=P.
Trivial.
Trivial.
Intros.
Apply H8.
Trivial.
Trivial.
Trivial.
Trivial.
Qed.


Theorem u_leads_to_symmetric_double_induction : 
  (IH:u_prog->u_prop->u_prop->u_prop->u_prop->Prop)
  (pr:u_prog)(u_prog_not_empty pr)->
  ((p,q,p',q':u_prop)(IH pr p q p' q')->(IH pr p' q' p q))->
  ((p,q,p',q':u_prop)
   (u_ensures pr p q)->(u_leads_to pr p' q')->(IH pr p q p' q'))->
  ((p,q,r,p',q',r':u_prop)
   ((u_ensures pr p q)->(u_leads_to pr q r)->
    (u_ensures pr p' q')->(u_leads_to pr q' r')->
    (IH pr p q p' q')->
    (IH pr p q q' r')->(IH pr q r p' q')->
    (IH pr q r q' r')->
    (IH pr p r q' r')->(IH pr q r p' r')->
    (IH pr p r p' r')))->
  ((T:Set)(P:T->u_prop)(p,q,p',q':u_prop)
    ((i:T)(u_leads_to pr (P i) q))->
    (u_leads_to pr p' q')->
    ((i:T)(IH pr (P i) q p' q'))->
    (u_everywhere (u_iff p (u_ex ? P)))->
    (IH pr p q p' q'))->  
  ((p,q,p',q':u_prop)
   (u_leads_to pr p q)->(u_leads_to pr p' q')->(IH pr p q p' q')).

Intros.
Apply u_leads_to_double_induction.
Trivial.
Trivial.
Intros.
Apply H0.
Apply H1.
Trivial.
Trivial.
Trivial.
Intros.
Apply H0.
Apply H3 with P:=P'.
Trivial.
Trivial.
Intros.
Apply H0.
Trivial.
Trivial.
Intros.
Apply H3 with P:=P.
Intros.
Trivial.
Trivial.
Trivial.
Trivial.
Trivial.
Trivial.
Qed.




Lemma u_leads_to_simple_conjunction_basis : (pr:u_prog)(u_prog_not_empty pr)->
  (p,q,p',q',r:u_prop)
  (u_ensures pr p q)->
  (u_leads_to pr p' q')->
  (u_unless pr q r)->
  (u_unless pr q' r)->
  (u_leads_to pr (u_and p p') (u_or (u_and q q') r)).

Intros.
Cut (u_ensures pr p q).
Intros E.
Unfold u_ensures in H0.
Elim H0.
Intros.
Clear H0.
Specialize (u_unless_cancelation pr p q r H4 H2).
Intros.
Specialize (u_leads_to_psp pr H ? ? ? ? H1 H0).
Intros.
Cut (u_leads_to pr (u_and p p') (u_or (u_and q' (u_or p q)) r)).
Intros.
Specialize (u_ensures_psp pr H ? ? ? ? E H3).
Intros.
Specialize (u_leads_to_basis pr ? ? H8).
Intros.
Clear H8.
Generalize (u_leads_to_implication pr H (u_and p p') (u_and p p')
             (u_or (u_and q' (u_or p q)) r)
             (u_or (u_or r (u_and q q')) (u_and p q'))).
Intros.
Cut (u_leads_to pr (u_and p p') (u_or (u_and q' (u_or p q)) r))
     ->(u_leads_to pr (u_and p p')
         (u_or (u_or r (u_and q q')) (u_and p q'))).
Intros.
Specialize (H10 H7).
Intros.
Clear H10.
Specialize (u_leads_to_cancelation pr H ? ? ? ? H11 H9).
Intros.
Apply (u_leads_to_implication_r pr H) with 2:=H10.
Unfold u_everywhere u_imp u_or u_and.
Clear H10 H8 H11 H9 H6 H7 H0 H5 H4 H3 E H2 H1.
Tauto.
Intros.
Apply H8.
Apply u_imp_p_p.
Unfold u_everywhere u_imp u_or u_and.
Clear H10 H8 H9 H6 H7 H0 H5 H4 E H3 H2 H1.
Tauto.
Apply (u_leads_to_implication_l pr H) with q:=(u_and p' (u_or p q)).
Unfold u_everywhere u_imp u_and u_or.
Clear H10 H8 H9 H6 H7 H0 H5 H4 E H3 H2 H1.
Tauto.
Trivial.
Apply (u_leads_to_implication_l pr H) with q:=(u_and p' (u_or p q)).
Unfold u_everywhere u_imp u_and u_or.
Clear H6 H0 H5 H4 E H3 H2 H1.
Tauto.
Trivial.
Trivial.
Qed.


Theorem u_leads_to_simple_conjunction : (pr:u_prog)(u_prog_not_empty pr)->
  (p,q,p',q',r:u_prop)
  (u_leads_to pr p q)->
  (u_leads_to pr p' q')->
  (u_unless pr q r)->
  (u_unless pr q' r)->
  (u_leads_to pr (u_and p p') (u_or (u_and q q') r)).

Intro.
Intro.
Generalize (u_leads_to_symmetric_double_induction
             [ppr:u_prog]
              [p,q,p',q':u_prop]
               (r:u_prop)
                (u_unless pr q r)
                 ->(u_unless pr q' r)
                    ->(u_leads_to pr (u_and p p') (u_or (u_and q q') r))
             pr H).
Intros.
Apply H0.
Intros.
Specialize (H5 ? H7 H6).
Intros.
Apply u_leads_to_implication with 4:=H8.
Trivial.
Unfold u_everywhere u_imp u_and u_and.
Intros.
Elim H9.
Intros.
Auto.
Unfold u_everywhere u_imp u_or u_and.
Intros.
Elim H9.
Intros.
Elim H10.
Auto.
Auto.
Intros.
Apply u_leads_to_simple_conjunction_basis.
Trivial.
Trivial.
Trivial.
Trivial.
Trivial.
Intros.
Clear H0.
Clear H9 H10 H11.
Specialize (H12 r1 H15 H16).
Specialize (H13 r1 H15 H16).
Specialize (H14 r1 H15 H16).
Clear H12 H13 H14.
Intros.
Cut (u_unless pr p0 q0).
Intros.
Specialize (u_ensures_conjunction pr p0 p'0 q0 q'0 H11 H7).
Intros.
Specialize (u_leads_to_basis pr ? ? H12).
Intros.
Clear H12.
Cut (u_leads_to pr (u_and p'0 q0) (u_or (u_and r0 r') r1)).
Intros.
Specialize (u_leads_to_ternary_disjunction pr H (u_and p0 q'0)
             (u_and p'0 q0) (u_and q0 q'0) ? H9 H12 H10).
Intros.
Apply u_leads_to_transitivity with 1:=H13 2:=H14.

Apply u_leads_to_implication_l with 3:=H0.
Trivial.
Unfold u_everywhere u_imp u_and.
Intros.
Elim H12.
Intros.
Auto.
Unfold u_ensures in H5.
Elim H5.
Auto.
Intros.
Clear H0.
Cut (i:T)(u_leads_to pr (u_and (P i) p'0) (u_or (u_and q0 q'0) r0)).
Intros.
Clear H7.
Clear H1 H2 H3 H4.
Apply (u_leads_to_disjunction pr T [i:T](u_and (P i) p'0)).
Trivial.
Unfold u_everywhere u_iff u_and u_ex.
Unfold u_everywhere u_iff u_ex in H8.
Intros.
Specialize (H8 s).
Intros.
Elim H1.
Intros.
Split.
Intros.
Elim H4.
Intros.
Elim (H2 H7).
Intros.
Split with x.
Auto.
Intros.
Split.
Apply H3.
Elim H4.
Intros.
Split with x.
Elim H7.
Auto.
Elim H4.
Intros.
Elim H7.
Auto.
Intros.
Apply H7.
Trivial.
Trivial.
Trivial.
Trivial.
Trivial.
Trivial.
Qed.


Theorem u_leads_to_simple_completion : (pr:u_prog)(u_prog_not_empty pr)->
  (p,q,p',q',r:u_prop)
  (u_leads_to pr p (u_or q r))-> 
  (u_leads_to pr p' (u_or q' r))->
  (u_unless pr q r)->
  (u_unless pr q' r)->
  (u_leads_to pr (u_and p p') (u_or (u_and q q') r)).

Intros.
Specialize (u_unless_refl pr r).
Intros.
Generalize (u_unless_simple_disjunction pr ? ? ? ? H2 H4).
Intro.
Generalize (u_unless_simple_disjunction pr ? ? ? ? H3 H4).
Intros.
Generalize (u_leads_to_simple_conjunction pr H ? ? ? ? ? H0 H1 H5 H6).
Intros.
Apply u_leads_to_implication_r with 3:=H7.
Trivial.
Unfold u_everywhere u_and u_or u_imp.
Intros.
Clear H0 H1 H2 H3 H4 H5 H6 H7.
Tauto.
Qed.


Theorem u_leads_to_completion_nat : (pr:u_prog)(u_prog_not_empty pr)->
  (b:nat)(P,Q:nat->u_prop)(r:u_prop)
  ((i:nat)(lt i b)->(u_leads_to pr (P i) (u_or (Q i) r)))->
  ((i:nat)(lt i b)->(u_unless pr (Q i) r))->
  (u_leads_to pr (u_all_nat b P) (u_or (u_all_nat b Q) r)).

Intros.
Generalize H0.
Generalize H1.
Clear H0 H1.
Generalize b.
Clear b.
Induction b.
Intros.
Intros.
Unfold u_ensures.
Apply (u_leads_to_implication_r pr H) with q:=u_true.
Unfold u_everywhere u_imp u_true u_or u_all_nat.
Intros.
Left .
Intros.
Elim (lt_n_O i H3).
Apply u_leads_to_true.
Trivial.
Intros.
Cut (i:nat)(lt i n)->(u_unless pr (Q i) r).
Intros.
Cut (i:nat)(lt i n)->(u_leads_to pr (P i) (u_or (Q i) r)).
Intros.
Specialize (H0 H3 H4).
Intros.
Clear H0.
Cut (u_leads_to pr (P n) (u_or (Q n) r)).
Intros.
Cut (u_unless pr (u_all_nat n Q) r).
Intros.
Cut (u_unless pr (Q n) r).
Intros.
Generalize (u_leads_to_simple_completion pr H ? ? ? ? ? H5 H0 H6 H7).
Intros.
Apply u_leads_to_implication with 4:=H8.
Trivial.
Unfold u_everywhere u_imp u_all_nat u_and.
Intros.
Generalize (H9 n).
Intros.
Split.
Intros.
Apply H9.
Apply lt_S.
Trivial.
Apply H10.
Apply lt_n_Sn.
Unfold u_everywhere u_imp u_or u_and u_all_nat.
Intros.
Elim H9.
Intros.
Left .
Intros.
Elim H10.
Intros.
Specialize (lt_n_Sm_le i n H11).
Intros.
Elim (le_lt_or_eq i n H14).
Intros.
Apply H12.
Trivial.
Intros.
Rewrite H15.
Trivial.
Intros.
Right .
Trivial.
Apply H1.
Apply lt_n_Sn.
Apply u_unless_left_big_conjunction.
Trivial.
Trivial.
Apply H2.
Apply lt_n_Sn.
Intros.
Apply H2.
Apply lt_S.
Trivial.
Intros.
Apply H1.
Apply lt_S.
Trivial.
Qed.


Definition card_le := [n:nat][X:Set] 
  (Ex [f:nat->X] ((x:X) (Ex [i:nat] (lt i n) /\ ((f i) = x)))).

Definition card_ge := [n:nat][X:Set]
  (Ex [f:X->nat] ((i:nat) (lt i n) -> (Ex [x:X] ((f x) = i)))).

Definition card_eq := [n:nat][X:Set]
  (Ex [f:nat->X] (Ex [g:X->nat] ((x:X) ((f (g x)) = x) /\ (lt (g x) n)))).

Definition finite := [X:Set] (Ex [n:nat] (card_le n X)).


Theorem u_leads_to_completion : (pr:u_prog)(u_prog_not_empty pr)->
  (T:Set)(finite T)->(P,Q:T->u_prop)(r:u_prop)
  ((i:T)(u_leads_to pr (P i) (u_or (Q i) r)))->
  ((i:T)(u_unless pr (Q i) r))->
  (u_leads_to pr (u_all T P) (u_or (u_all T Q) r)).

Intros pr NE T F P Q r.
Unfold finite in F.
Elim F.
Intro b.
Intro.
Clear F.
Unfold card_le in H.
Elim H.
Intro h.
Intro.
Clear H.
Intros.
Cut (u_leads_to pr (u_all_nat b [n:nat](P (h n)))
      (u_or (u_all_nat b [n:nat](Q (h n))) r)).
Intros.
Apply u_leads_to_implication with 4:=H2.
Trivial.
Unfold u_everywhere u_imp u_all u_all_nat.
Intros.
Apply H3.
Unfold u_everywhere u_imp u_or u_all_nat u_all.
Intros.
Elim H3.
Intros.
Left .
Intros.
Elim (H0 t).
Intros.
Elim H5.
Intros.
Rewrite <- H7.
Apply H4.
Trivial.
Intros.
Right .
Trivial.
Apply u_leads_to_completion_nat.
Trivial.
Intros.
Apply H.
Intros.
Apply H1.
Qed.



Theorem u_leads_to_induction : (pr:u_prog)(u_prog_not_empty pr)->
  (T:Set)(less:T->T->Prop)(variant:u_st->T)
  (well_founded T less)->
  (p,q:u_prop)
  ((m:T)(u_leads_to pr (u_and p [s:u_st] (variant s) = m)
                       (u_or (u_and p [s:u_st] (less (variant s) m)) q)))->
  (u_leads_to pr p q).

Intros.
Cut (m:T)(u_leads_to pr (u_and p [s:u_st](variant s)=m) q).
Intros.
Generalize (u_leads_to_disjunction_2 pr T
             [m:T](u_and p [s:u_st](variant s)=m) q).
Intros.
Generalize (H3 H2).
Intros.
Clear  H3.
Apply u_leads_to_implication_l with q:=(u_ex T
                                         [m:T]
                                          (u_and p [s:u_st](variant s)=m)).
Assumption.
Unfold u_everywhere u_imp u_ex u_and.
Intros.
Exists  (variant s).
Split.
Assumption.
Reflexivity.
Assumption.
Intros.
Apply well_founded_ind with A:=T R:=less P:=[m:T]
                                                          (u_leads_to
                                                            pr
                                                            (u_and p
                                                              [s:u_st]
                                                              (variant s)=m)
                                                            q).
Assumption.
Intros.
Generalize (u_leads_to_disjunction_2 pr {y:T | (less y x)}).
Intros.
Generalize (H3
             [i:{y:T | (less y x)}]
              (u_and p [s:u_st](variant s)=(proj1_sig ? ? i))).
Intros.
Generalize (H4 q).
Intros.
Clear  H3 H4.
Cut (u_leads_to pr
      (u_ex {y:T | (less y x)}
        [i:{y:T | (less y x)}]
         (u_and p
           [s:u_st]
            (variant s)
             =(proj1_sig T [y:T](less y x) i)))
      q).
Intros.
Cut (u_leads_to pr q q).
Intros.
Generalize (u_leads_to_small_disjunction pr H
             (u_ex {y:T | (less y x)}
               [i:{y:T | (less y x)}]
                (u_and p
                  [s:u_st]
                   (variant s)
                    =(proj1_sig T
                       [y:T](less y x) i))) q q q H3 H4).
Intros.
Apply u_leads_to_transitivity with q:=(u_or
                                        (u_ex
                                          {y:T | 
                                           (less y x)}
                                          [i:{y:T | 
                                              (less y x)}]
                                           (u_and p
                                             [s:u_st]
                                              (variant s)
                                               =(proj1_sig
                                                  T
                                                  [y:T]
                                                   (less y x) i))) q).
Clear  H5 H3 H4 H6.
Generalize (H1 x).
Clear  H1.
Intros.
Apply u_leads_to_implication_r with q:=(u_or
                                         (u_and p
                                           [s:u_st](less (variant s) x)) q).
Assumption.
Unfold u_everywhere u_imp u_or u_and u_ex.
Intros.
Elim H3.
Intros.
Elim H4.
Intros.
Left.
Exists  (exist T [y:T](less y x) (variant s) H6).
Split.
Assumption.
Reflexivity.
Intros.
Right.
Assumption.
Assumption.
Apply u_leads_to_implication_r with q:=(u_or q q).
Assumption.
Unfold u_everywhere u_imp u_or.
Intros.
Elim H7.
Auto.
Auto.
Assumption.
Apply u_leads_to_refl.
Assumption.
Clear  H5.
Apply (u_leads_to_disjunction_2 pr {y:T | (less y x)}).
Intros.
Elim i.
Intros y' H'.
Generalize (H2 y' H').
Intros.
Auto.
Qed.


Theorem u_leads_to_var_induction : (pr:u_prog)(u_prog_not_empty pr)->
  (M:u_var)(less:(u_var_type M)->(u_var_type M)->Prop)
  (well_founded (u_var_type M) less)->
  (p,q:u_prop)
  ((m:(u_var_type M))(u_leads_to pr (u_and p [s:u_st] (s M) = m)
                        (u_or (u_and p [s:u_st] (less (s M) m)) q)))->
  (u_leads_to pr p q).

Intros.
Apply (u_leads_to_induction pr H (u_var_type M) less [s:u_st](s M)).
Trivial.
Trivial.
Qed.


Theorem u_leads_to_nat_induction : (pr:u_prog)(u_prog_not_empty pr)->
  (variant:u_st->nat)
  (p,q:u_prop)
  ((m:nat)(u_leads_to pr (u_and p [s:u_st] (variant s) = m)
                         (u_or (u_and p [s:u_st] (lt (variant s) m)) q)))->
  (u_leads_to pr p q).

Intros pr H v p q.
Intros.
Apply (u_leads_to_induction pr H nat lt v lt_wf p q).
Auto.
Qed.


Theorem u_leads_to_rev_nat_induction : 
  (pr:u_prog)(u_prog_not_empty pr)->
  (variant:u_st->nat)(maximum:nat)(p,q:u_prop)
  (u_everywhere (u_imp p [s:u_st](le (variant s) maximum)))->
  ((m:nat)(le m maximum)->
     (u_leads_to pr (u_and p [s:u_st] m = (variant s))
                    (u_or (u_and p [s:u_st] (lt m (variant s))) q)))->
  (u_leads_to pr p q).

Intro.
Intro NE.
Intros.
Apply (u_leads_to_nat_induction pr NE
        [s:u_st](minus maximum (variant s)) p q).
Intros.
Elim (le_or_lt m maximum).
Intros L.
Cut (le (minus maximum m) maximum).
Intros L'.
Generalize (H0 (minus maximum m) L').
Intros.
Clear H0.
Apply u_leads_to_implication with 4:=H1.
Trivial.

Unfold u_everywhere u_imp u_and.
Intros.
Elim H0.
Clear H0.
Intros.
Split.
Assumption.

Generalize (H s H0).
Intros.
Generalize (le_plus_minus ? ? H3).
Intros.
Rewrite H4.
Rewrite H2.
Rewrite (plus_sym (variant s) m).
Apply minus_plus.

Unfold u_everywhere u_imp u_and u_or.
Intros.
Elim H0.
Clear H0.
Intros.
Elim H0.
Clear H0.
Intros.
Left .
Split.
Trivial.

Apply lt_minus_imp_lt_minus.
Trivial.

Apply (H s H0).

Trivial.

Auto.

Intros.
Apply le_minus.

Intros.
Apply (u_leads_to_implication_l pr NE
        (u_and p [s:u_st](minus maximum (variant s))=m) u_false
        (u_or (u_and p [s:u_st](lt (minus maximum (variant s)) m)) q)).
Unfold u_everywhere u_imp u_and u_false.
Intros.
Generalize (lt_imp_lt_minus_l (variant s) maximum m).
Intros.
Generalize (H3 H1).
Clear H3.
Intros.
Elim H2.
Clear H2.
Intros.
Rewrite H4 in H3.
Apply (lt_n_n ? H3).

Apply u_leads_to_false.
Trivial.
Qed.

