(*C*
      Embedding UNITY into the Calculus of Inductive Constructions
                                     Using the COQ Proof Assistent
      Author:  Mark-Oliver Stehr
               University of Hamburg
               Department of Computer Science
               Theoretical Foundations of Computer Science Group 
               stehr@informatik.uni-hamburg.de
      Date:    September 1998
*C*)




Definition u_prop := (u_st->Prop).


Definition u_true := [p:u_st] True.
Definition u_false := [p:u_st] False.
Definition u_or := [p,q:u_prop][s:u_st] (p s) \/ (q s).
Definition u_or3 := [p,q,r:u_prop][s:u_st] (p s) \/ (q s) \/ (r s).
Definition u_and := [p,q:u_prop][s:u_st] (p s) /\ (q s).
Definition u_and3 := [p,q,r:u_prop][s:u_st] (p s) /\ (q s) /\ (r s).
Definition u_imp := [p,q:u_prop][s:u_st] (p s) -> (q s).
Definition u_iff := [p,q:u_prop][s:u_st] (p s) <-> (q s).
Definition u_not := [p:u_prop][s:u_st] ~(p s).
Definition u_ex := [T:Set][P:T->u_prop] [s:u_st] (Ex [t:T] ((P t) s)).
Definition u_all := [T:Set][P:T->u_prop] [s:u_st] (t:T)((P t) s).
Definition u_all_nat := [b:nat][P:nat->u_prop][s:u_st]
                        ((i:nat)(lt i b)->((P i) s)).

Definition u_everywhere := [p:u_prop] (s:u_st)(p s).



Theorem u_imp_p_p : (p:u_prop)(u_everywhere (u_imp p p)).

Unfold u_everywhere u_imp.
Intros.
Assumption.
Qed.

Theorem u_imp_false : (p:u_prop)(u_everywhere (u_imp u_false p)).

Intros.
Unfold u_everywhere u_imp u_false.
Intros.
Elim H.
Qed.

Theorem u_imp_true : (p:u_prop)(u_everywhere (u_imp p u_true)).

Intros.
Unfold u_everywhere u_imp u_true.
Intros.
Exact I.
Qed.

Theorem u_iff_imp : (p,q:u_prop)
  (u_everywhere (u_iff p q))->(u_everywhere (u_imp p q)).

Unfold u_everywhere u_iff u_imp.
Intros.
Specialize (H s).
Intro.
Elim H1.
Intros.
Apply H2.
Assumption.
Qed.

Theorem u_iff_refl : (p:u_prop)(u_everywhere (u_iff p p)).

Unfold u_everywhere u_iff.
Intros.
Split.
Intros.
Exact H.
Intros.
Exact H.
Qed.

Theorem u_iff_sym : (p,q:u_prop)
  (u_everywhere (u_iff p q))->(u_everywhere (u_iff q p)).

Unfold u_everywhere.
Unfold u_iff.
Intros.
Specialize (H s).
Intro.
Elim H0.
Intros.
Split.
Assumption.
Assumption.
Qed.

Theorem u_iff_trans : (p,q,r:u_prop)
  (u_everywhere (u_iff p q))->(u_everywhere (u_iff q r))->
  (u_everywhere (u_iff p r)).

Unfold u_everywhere u_iff.
Intros.
Specialize (H s).
Specialize (H0 s).
Intros.
Split.
Elim H1.
Elim H2.
Intros.
Auto.
Elim H1.
Elim H2.
Intros.
Auto.
Qed.

Theorem u_imp_and_l : (p,q:u_prop)(u_everywhere (u_imp (u_and p q) p)).

Unfold u_everywhere u_imp u_and.
Intros.
Elim H.
Intros.
Assumption.
Qed.

Theorem u_imp_and_r : (p,q:u_prop)(u_everywhere (u_imp (u_and p q) q)).

Unfold u_everywhere u_imp u_and.
Intros.
Elim H.
Intros.
Assumption.
Qed.

Theorem u_imp_or_l : (p,q:u_prop)(u_everywhere (u_imp p (u_or p q))).

Unfold u_everywhere u_imp u_or.
Intros.
Left.
Assumption.
Qed.

Theorem u_imp_or_r : (p,q:u_prop)(u_everywhere (u_imp q (u_or p q))).

Unfold u_everywhere u_imp u_or.
Intros.
Right.
Assumption.
Qed.




Axiom classic : (p:Prop)(p \/ ~p).


Theorem u_classic : (p:u_prop)(u_everywhere (u_or p (u_not p))).

Unfold u_everywhere u_or u_not.
Unfold u_everywhere u_or u_not.
Intros.
Apply classic.
Qed.






