(*C*
      Embedding UNITY into the Calculus of Inductive Constructions
                                     Using the COQ Proof Assistent
      Author:  Mark-Oliver Stehr
               University of Hamburg
               Department of Computer Science
               Theoretical Foundations of Computer Science Group 
               stehr@informatik.uni-hamburg.de
      Date:    September 1998
*C*)





Definition u_prog := (list u_tr).





Definition u_prog_not_empty := [pr:u_prog] (Ex [t:u_tr] (In t pr)).



Fixpoint prog_contains [pr:(list u_tr)] : u_tr-> bool :=
  [stmt:u_tr] Case pr of
    (* (nil u_tr) *) false
    (* (cons h t) *) [h:u_tr][t:(list u_tr)]
      Case (u_tr_eq_dec stmt h) of
        (* (stmt=h) *) [_] true
        (* ~(stmt=h) *) [_] (prog_contains t stmt)
      end
  end.


Theorem prog_contains_spec : (pr:u_prog)(t:u_tr)
  ((prog_contains pr t)=true)<->(In t pr).

Induction pr.
Intro.
Split.
Simpl.
Intro.
Apply diff_true_false.
Symmetry.
Assumption.
Simpl.
Intro.
Apply False_ind.
Assumption.
Intros.
Simpl.
Elim (u_tr_eq_dec t a).
Split.
Intro.
Left.
Symmetry.
Assumption.
Intro.
Reflexivity.
Intro.
Split.
Intro.
Right.
Generalize (H t).
Intro.
Elim H1.
Intros.
Apply H2.
Assumption.
Intro.
Elim H0.
Intro.
Apply False_ind.
Apply (y (sym_equal u_tr a t H1)).
Generalize (H t).
Intro.
Elim H1.
Intros.
Apply H3.
Assumption.
Qed.


Theorem u_prog_contains : (pr:u_prog)(t:u_tr){(In t pr)}+{~(In t pr)}.

Intros.
Cut {(prog_contains pr t)=true}+{(prog_contains pr t)=false}.
Intros.
Elim H.
Intro.
Left.
Elim (prog_contains_spec pr t).
Intros.
Apply H0.
Assumption.
Intro.
Elim (prog_contains_spec pr t).
Intros.
Right.
Unfold not.
Intro.
Generalize (H1 H2).
Intro.
Rewrite y in H3.
Apply diff_true_false.
Symmetry.
Assumption.
Elim (prog_contains pr t).
Left.
Reflexivity.
Right.
Reflexivity.
Qed.


Fixpoint u_prog_union [pr:(list u_tr)] : (list u_tr)->(list u_tr) :=
[pr':(list u_tr)] Case pr of
  (* (nil u_tr) *) pr'
  (* (cons h t) *) [h:u_tr][t:(list u_tr)]
    Case (u_prog_contains pr' h) of
      (* (prog_contains pr' h) *) [_](u_prog_union t pr')
      (* ~(prog_contains pr' h) *) [_](u_prog_union t (cons h pr'))
    end
  end.


Theorem u_prog_union_spec : (pr,pr':(list u_tr))(t:u_tr)
  ((In t pr) \/ (In t pr')) <-> (In t (u_prog_union pr pr')).

Induction pr.
Simpl.
Intros.
Split.
Intros.
Elim H.
Intros.
Elim H0.
Auto.
Intros.
Right.
Assumption.
Intros.
Simpl.
Elim (u_prog_contains pr' a).
Split.
Intros.
Elim H0.
Intros.
Elim H1.
Intros.
Rewrite H2 in y.
Generalize (H pr' t).
Intros.
Elim H3.
Intros.
Apply H4.
Right.
Assumption.
Intro.
Generalize (H pr' t).
Intros.
Elim H3.
Intros.
Apply H4.
Left.
Assumption.
Intro.
Generalize (H pr' t).
Intros.
Elim H2.
Intros.
Apply H3.
Right.
Assumption.
Intros.
Generalize (H pr' t).
Intros.
Elim H1.
Intros.
Generalize (H3 H0).
Intros.
Elim H4.
Intros.
Left.
Right.
Assumption.
Intros.
Right.
Assumption.
Intros.
Generalize (H (cons a pr') t).
Intros.
Elim H0.
Intros.
Split.
Intros.
Apply H1.
Elim H3.
Intros.
Elim H4.
Intros.
Right.
Rewrite -> H5.
Simpl.
Left.
Reflexivity.
Intros.
Left.
Assumption.
Intros.
Right.
Simpl.
Right.
Assumption.
Intros.
Generalize (H2 H3).
Intros.
Elim H4.
Intros.
Left.
Right.
Assumption.
Simpl.
Intros.
Elim H5.
Intros.
Left.
Left.
Assumption.
Intros.
Right.
Assumption.
Qed.


Theorem u_prog_union_spec_1 : (pr,pr':(list u_tr))(t:u_tr)
  ((In t pr) \/ (In t pr')) -> (In t (u_prog_union pr pr')).

Intros.
Elim (u_prog_union_spec pr pr' t).
Intros.
Apply H0.
Trivial.
Qed.

Theorem u_prog_union_spec_2 : (pr,pr':(list u_tr))(t:u_tr)
  (In t (u_prog_union pr pr')) -> ((In t pr) \/ (In t pr')). 

Intros.
Elim (u_prog_union_spec pr pr' t).
Intros.
Apply H1.
Trivial.
Qed.







