(*C*
      Embedding UNITY into the Calculus of Inductive Constructions
                                     Using the COQ Proof Assistent
      Author:  Mark-Oliver Stehr
               University of Hamburg
               Department of Computer Science
               Theoretical Foundations of Computer Science Group 
               stehr@informatik.uni-hamburg.de
      Date:    September 1998
*C*)




Fixpoint u_co [pr:u_prog] : u_prop->u_prop->Prop :=
  [p,q:u_prop] Cases pr of
                 nil => True
               | (cons h t) => ((s:u_st)(p s)->(q ((u_effect h) s))) /\ 
                               (u_co t p q) 
               end.



Theorem u_co_characterization : (pr:u_prog)(p,q:u_prop)
  (u_co pr p q) <->
  ((t:u_tr)(s:u_st)(In t pr)->(p s)->(q ((u_effect t) s))).

Induction pr.
Simpl.
Split.
Intros.
Apply False_ind.
Exact H0.
Intros.
Exact I.
Intros.
Simpl.
Generalize (H p q).
Clear  H.
Intro.
Elim H.
Intros.
Clear  H.
Split.
Intros.
Elim H2.
Intro.
Clear  H2.
Elim H.
Intros.
Rewrite <- H4.
Apply H2.
Assumption.
Intro.
Elim H.
Intros.
Generalize (H0 H6 t s).
Intros.
Apply H7.
Assumption.
Assumption.
Intros.
Split.
Intros.
Apply (H a s).
Constructor 1.
Reflexivity.
Assumption.
Clear  H0.
Apply H1.
Intros.
Apply (H t s).
Constructor 2.
Assumption.
Assumption.
Qed.


Theorem u_nc_co : (pr:u_prog)(p,q:u_prop)
  (u_co pr p q) ->
  ((t:u_tr)(s:u_st)(In t pr)->(p s)->(q ((u_effect t) s))).

Intros pr p q.
Intros.
Apply (proj1 ? ? (u_co_characterization pr p q)).
Assumption.
Assumption.
Assumption.
Qed.

Theorem u_sc_co : (pr:u_prog)(p,q:u_prop)
  ((t:u_tr)(s:u_st)(In t pr)->(p s)->(q ((u_effect t) s)))->
  (u_co pr p q).

Intros pr p q.
Intros.
Apply (proj2 ? ? (u_co_characterization pr p q)).
Assumption.
Qed.




Definition u_unless := [pr:u_prog][p,q:u_prop] 
  (u_co pr (u_and p (u_not q)) (u_or p q)).




Definition u_stable := [pr:u_prog][p:u_prop] (u_co pr p p).


Definition u_ind_invariant := [pr:u_prog][initially:u_prop][p:u_prop] 
  (u_everywhere (u_imp initially p)) /\ (u_stable pr p).


Definition u_invariant := [pr:u_prog][initially:u_prop][p:u_prop]
  (ExT [inv:u_prop] (u_ind_invariant pr initially inv) /\
                    (u_everywhere (u_imp inv p))).


Definition u_si := [pr:u_prog][initially:u_prop][s:u_st]
  (p:u_prop)(u_ind_invariant pr initially p)->(p s).



Theorem u_co_implication : (pr:u_prog)(p,p',q,q':u_prop)
  (u_everywhere (u_imp p p')) ->
  (u_everywhere (u_imp q q')) ->
  (u_co pr p' q) -> (u_co pr p q').

Unfold u_everywhere u_imp.
Intros.
Apply u_sc_co.
Intros.
Apply H0.
Apply (u_nc_co pr p' q).
Assumption.
Assumption.
Apply H.
Assumption.
Qed.

Theorem u_co_implication_r : (pr:u_prog)(p,q,r:u_prop)
  (u_everywhere (u_imp q r)) ->
  (u_co pr p q) -> (u_co pr p r).

Intros.
Apply (u_co_implication pr p p q r).
Apply u_imp_p_p.
Assumption.
Assumption.
Qed.

Theorem u_co_implication_l : (pr:u_prog)(p,q,r:u_prop)
  (u_everywhere (u_imp p q)) ->
  (u_co pr q r) -> (u_co pr p r).

Intros.
Apply (u_co_implication pr p q r r).
Assumption.
Apply u_imp_p_p.
Assumption.
Qed.


Theorem u_co_subst_r : (pr:u_prog)(p,q,q':u_prop)
  (u_everywhere (u_iff q q'))->(u_co pr p q)->(u_co pr p q').

Intros.
Apply u_co_implication_r with q:=q.
Red.
Red in H.
Intros.
Specialize (H s).
Intros.
Unfold u_imp.
Unfold u_iff in H1.
Elim H1.
Intros H2 H3.
Exact H2.
Exact H0.
Qed.

Theorem u_co_subst_l : (pr:u_prog)(p,p',q:u_prop)
  (u_everywhere (u_iff p p'))->(u_co pr p q)->(u_co pr p' q).

Intros.
EApply u_co_implication_l.
Unfold u_everywhere.
Unfold u_everywhere in H.
Unfold u_imp.
Unfold u_iff in H.
Intros.
Specialize (H s).
Intros.
Unfold iff in H2.
Elim H2.
Intros.
Apply H4.
Exact H1.
Exact H0.
Qed.



Theorem u_co_conjunction : (pr:u_prog)(p,p',q,q':u_prop)
  (u_co pr p q) -> (u_co pr p' q') -> (u_co pr (u_and p p') (u_and q q')).

Intros.
Apply u_sc_co.
Intros.
Unfold u_and.
Unfold u_and in H2.
Elim H2.
Intros.
Split.
Apply (u_nc_co pr p q).
Assumption.
Assumption.
Assumption.
Apply (u_nc_co pr p' q').
Assumption.
Assumption.
Assumption.
Qed.


Theorem u_stable_subst : (pr:u_prog)(p,q:u_prop)
  (u_everywhere (u_iff p q)) ->
  (u_stable pr p) -> (u_stable pr q).

Unfold u_stable.
Intros.
Apply (u_co_implication pr) with p':=p q:=p.
Apply u_iff_imp.
Apply u_iff_sym.
Auto.
Apply u_iff_imp.
Auto.
Auto.
Qed.

Theorem u_ind_invariant_subst : (pr:u_prog)(r,r',p,q:u_prop)
  (u_everywhere (u_imp r' r))->(u_everywhere (u_iff p q)) ->
  (u_ind_invariant pr r p) -> (u_ind_invariant pr r' q).

Unfold u_ind_invariant.
Intros.
Elim H1.
Clear H1.
Intros.
Split.
Unfold u_everywhere u_imp.
Unfold u_everywhere u_imp in H.
Unfold u_everywhere u_iff in H0.
Unfold u_everywhere u_imp in H1.
Intros.
Generalize (H0 s).
Intros.
Elim H4.
Intros.
Apply H5.
Apply H1.
Apply H.
Trivial.
Apply u_stable_subst with 1:=H0.
Trivial.
Qed.


Theorem u_co_stable_conjunction : (pr:u_prog)(p,q,r:u_prop)
  (u_stable pr r) -> (u_co pr p q) -> 
  (u_co pr (u_and p r) (u_and q r)).

Intros.
Unfold u_stable in H.
Apply (u_co_conjunction pr p r q r).
Auto.
Auto.
Qed.


Theorem u_rel_co : (pr:u_prog)(p,q,r:u_prop)
  (u_stable pr r)->(u_co pr (u_and r p) (u_imp r q))->
  (u_co pr (u_and r p) (u_and r q)).

Intros.
Generalize (u_co_stable_conjunction ? ? ? ? H H0).
Intros.
Clear H0.
Apply u_co_implication with 3:=H1.
Clear H H1.
Unfold u_everywhere u_imp u_and.
Tauto.
Unfold u_everywhere u_imp u_and.
Tauto.
Qed.

Theorem u_rel_stable : (pr:u_prog)(p,r:u_prop)
  (u_stable pr r)->(u_co pr (u_and r p) (u_imp r p))->
  (u_stable pr (u_and r p)).

Intros.
Unfold u_stable.
Apply u_rel_co.
Trivial.
Apply u_co_implication_r with 2:=H0.
Clear H H0.
Unfold u_everywhere u_imp u_and.
Tauto.
Qed.



Theorem u_ind_inv_is_inv : (pr:u_prog)(initially:u_prop)(p:u_prop)
  (u_ind_invariant pr initially p)->(u_invariant pr initially p).

Intros.
Unfold u_invariant.
Split with p.
Split.
Trivial.
Unfold u_everywhere u_imp.
Auto.
Qed.

Theorem u_si_is_indinv : (pr:u_prog)(initially:u_prop)
  (u_ind_invariant pr initially (u_si pr initially)).

Unfold u_si.
Intros.
Unfold u_ind_invariant.
Split.
Unfold u_everywhere u_imp.
Intros.
Elim H0.
Intros.
Apply H1.
Trivial.
Unfold u_stable u_everywhere u_imp.
Apply u_sc_co.
Intros.
Generalize (H0 p).
Clear H0.
Intros.
Cut (p s).
Intros.
Elim H1.
Intros.
Generalize (u_nc_co ? ? ? H4).
Intros.
Apply H5.
Trivial.
Trivial.
Apply H0.
Trivial.
Qed.

Theorem u_si_is_inv : (pr:u_prog)(initially:u_prop)
  (u_invariant pr initially (u_si pr initially)).

Intros.
Unfold u_invariant.
Split with (u_si pr initially).
Split.
Apply u_si_is_indinv.
Unfold u_everywhere u_imp.
Tauto.
Qed.

Theorem u_indinv_imp_si : (pr:u_prog)(initially:u_prop)(p:u_prop)
  (u_ind_invariant pr initially p)->
  (u_everywhere (u_imp (u_si pr initially) p)).

Unfold u_everywhere u_imp u_si.
Intros.
Apply H0.
Apply u_ind_invariant_subst with 3:=H.
Unfold u_everywhere u_imp.
Auto.
Unfold u_everywhere u_iff.
Clear H H0.
Tauto.
Qed.

Theorem u_inv_imp_si : (pr:u_prog)(initially:u_prop)(p:u_prop)
  (u_invariant pr initially p)->
  (u_everywhere (u_imp (u_si pr initially) p)).

Intros.
Elim H.
Clear H.
Intros.
Elim H.
Clear H.
Intros.
Cut (u_everywhere (u_imp (u_si pr initially) x)).
Generalize H0.
Unfold u_everywhere u_imp.
Clear H H0.
Intros.
Apply H0.
Apply H.
Trivial.
Apply u_indinv_imp_si.
Trivial.
Qed.


Theorem u_si_imp_inv : (pr:u_prog)(initially:u_prop)(p:u_prop)
  (u_everywhere (u_imp (u_si pr initially) p))->
  (u_invariant pr initially p).

Intros.
Unfold u_invariant.
Split with (u_si pr initially).
Split.
Apply u_si_is_indinv.
Trivial.
Qed.

Theorem u_inv_char : (pr:u_prog)(initially:u_prop)(p:u_prop)
  (u_invariant pr initially p)<->
  (u_everywhere (u_imp (u_si pr initially) p)).

Intros.
Split.
Intros.
Apply u_inv_imp_si.
Trivial.
Intros.
Apply u_si_imp_inv.
Trivial.
Qed.


Theorem u_rel_inv : (pr:u_prog)(initially,p,r:u_prop)
  (u_ind_invariant pr initially r)->
  (u_everywhere (u_imp (u_and initially r) p))->
  (u_co pr (u_and r p) (u_imp r p))->
  (u_ind_invariant pr initially (u_and r p)).

Intros.
Unfold u_ind_invariant.
Split.
Unfold u_ind_invariant in H.
Elim H.
Intros.
Clear H3 H1 H.
Unfold u_everywhere u_imp u_and.
Unfold u_everywhere u_imp u_and in H0.
Unfold u_everywhere u_imp in H2.
Intros.
Split.
Apply H2.
Trivial.
Apply H0.
Split.
Trivial.
Apply H2.
Trivial.
Apply u_rel_stable.
Unfold u_ind_invariant in H.
Elim H.
Intros.
Trivial.
Trivial.
Qed.


Theorem u_unless_stable_conjunction : (pr:u_prog)(p,q,r:u_prop)
  (u_stable pr r) -> (u_unless pr p q) -> 
  (u_unless pr (u_and p r) (u_and q r)).

Unfold u_unless.
Intros.
Generalize (u_co_stable_conjunction pr (u_and p (u_not q)) (u_or p q) r
             H H0).
Intros.
Clear  H0.
Apply u_co_implication with p':=(u_and (u_and p (u_not q)) r) q:=(
u_and (u_or p q) r).
Unfold u_everywhere u_imp u_and u_not.
Intros.
Split.
Elim H0.
Intros.
Split.
Elim H2.
Intros.
Auto.
Unfold not.
Intros.
Elim H3.
Split.
Auto.
Elim H2.
Intros.
Auto.
Elim H0.
Intros.
Elim H2.
Intros.
Auto.
Unfold u_everywhere u_imp u_and u_or.
Intros.
Elim H0.
Intros.
Elim H2.
Intros.
Left.
Split.
Auto.
Auto.
Intros.
Right.
Split.
Auto.
Auto.
Auto.
Qed.


Theorem u_rel_unless : (pr:u_prog)(p,q,r:u_prop)
  (u_stable pr r)->(u_co pr (u_and3 r p (u_not q)) (u_imp r (u_or p q)))->
  (u_unless pr (u_and r p) (u_and r q)).

Intros.
Unfold u_unless.
Generalize (u_co_stable_conjunction ? ? ? ? H H0).
Clear H H0.
Intros.
Apply u_co_implication with 3:=H.
Clear H.
Unfold u_everywhere u_imp u_and u_and3.
Unfold u_not.
Tauto.
Unfold u_everywhere u_imp u_and u_or.
Tauto.
Qed.


Theorem u_co_big_conjunction : (T:Set)(pr:u_prog)(P,Q:T->u_prop)
  ((t:T)(u_co pr (P t) (Q t)))->(u_co pr (u_all ? P) (u_all ? Q)).

Intros.
Apply u_sc_co.
Intros.
Unfold u_all.
Unfold u_all in H1.
Intros.
Apply u_nc_co with pr:=pr p:=(P t0).
Specialize (H t0). 
Intros.
Apply u_co_implication_r with pr:=pr q:=(Q t0).
Unfold u_everywhere.
Unfold u_imp.
Intros.
Tauto.
Exact H2.
Exact H0.
Exact (H1 t0).
Qed.


Theorem u_co_big_disjunction : (T:Set)(pr:u_prog)(P,Q:T->u_prop)
  ((t:T)(u_co pr (P t) (Q t)))->(u_co pr (u_ex ? P) (u_ex ? Q)).

Intros.
Apply u_sc_co.
Intros.
Unfold u_ex in H1.
Unfold u_ex.
Elim H1.
Intros t'.
Intro.
Exists  t'.
Generalize (H t').
Intro.
Apply (u_nc_co pr (P t') (Q t')).
Assumption.
Assumption.
Assumption.
Qed.

Theorem u_co_disjunction : (pr:u_prog)(p,p',q,q':u_prop)
 (u_co pr p q) -> (u_co pr p' q') -> (u_co pr (u_or p p') (u_or q q')).

Intros.
Generalize (u_co_big_disjunction bool pr).
Intros.
Generalize (H1 [x:bool](bool_rect [_:bool]u_prop p p' x)).
Intros.
Generalize (H2 [x:bool](bool_rect [_:bool]u_prop q q' x)).
Intros.
Apply u_co_implication
 with p':=(u_ex bool [x:bool](bool_rect [_:bool]u_prop p p' x))
      q:=(u_ex bool [x:bool](bool_rect [_:bool]u_prop q q' x)).
Red.
(Intros; Red).
Intros.
Red.
Red in H4.
Elim H4.
Intro.
Split with true.
Red.
Trivial.
(Intro; Split with false; Red; Trivial).
Red.
(Unfold u_imp; Intros).
Red in H4.
Elim H4.
Induction x.
Intro.
Red.
Red in H5.
(Left ; Assumption).
Intro.
Red in H5.
Red.
Right .
Assumption.
Clear H2.
Apply H3.
Clear H3.
Intros.
Clear H1.
Elim t.
Simpl.
Trivial.
(Simpl; Assumption).
Qed.



Theorem u_co_false : (pr:u_prog)(p:u_prop)(u_co pr u_false p).

Intros.
Apply u_sc_co.
Intros.
Unfold u_false in H0.
Elim H0.
Qed.

Theorem u_co_true : (pr:u_prog)(p:u_prop)(u_co pr p u_true).

Intros.
Apply u_sc_co.
Intros.
Unfold u_true.
Try Trivial.
Qed.

Theorem u_unless_true : (pr:u_prog)(p:u_prop)(u_unless pr p u_true).

Unfold u_unless.
Intros.
Specialize (u_co_true pr (u_and p (u_not u_true))).
Intros.
Apply u_co_implication_r with 2:=H.
Unfold u_everywhere u_imp u_or.
Unfold u_true.
Tauto.
Qed.



Theorem u_unless_subst_r : (pr:u_prog)(p,q,q':u_prop)
  (u_everywhere (u_iff q q'))->(u_unless pr p q)->(u_unless pr p q').

Unfold u_unless.
Intros.
Apply u_co_implication_r with q:=(u_or p q).
Unfold u_everywhere u_imp u_or u_or.
Unfold u_everywhere u_iff in H.
Intros.
Elim (H s).
Intros.
Elim H1.
Intros.
Left.
Assumption.
Intros.
Right.
Apply H2.
Assumption.
Apply u_co_implication_l with q:=(u_and p (u_not q)).
Unfold u_everywhere u_imp u_and u_not.
Unfold u_everywhere u_iff in H.
Intros.
Elim (H s).
Intros.
Elim H1.
Intros.
Split.
Assumption.
Unfold not in H5.
Unfold not.
Intros.
Apply H5.
Apply H2.
Assumption.
Assumption.
Qed.

Theorem u_unless_subst_l : (pr:u_prog)(p,p',q:u_prop)
  (u_everywhere (u_iff p p'))->(u_unless pr p q)->(u_unless pr p' q).

Unfold u_unless.
Intros.
Apply u_co_implication_r with q:=(u_or p q).
Unfold u_everywhere u_imp u_or u_or.
Unfold u_everywhere u_iff in H.
Intros.
Elim (H s).
Intros.
Elim H1.
Intros.
Left.
Apply H2.
Assumption.
Intros.
Right.
Assumption.
Apply u_co_implication_l with q:=(u_and p (u_not q)).
Unfold u_everywhere u_imp u_and u_not.
Unfold u_everywhere u_iff in H.
Intros.
Elim (H s).
Intros.
Elim H1.
Intros.
Split.
Apply H3.
Assumption.
Assumption.
Assumption.
Qed.


Theorem u_unless_refl : (pr:u_prog)(p:u_prop)(u_unless pr p p).

Intros.
Unfold u_unless.
Apply u_co_implication_l with q:=u_false.
Unfold u_everywhere u_imp u_and u_not u_false.
Intros.
Elim H.
Intros.
Apply (H1 H0).
Apply u_co_false.
Qed.


Theorem u_unless_imp : (pr:u_prog)(p,q:u_prop)
  (u_everywhere (u_imp p q))->(u_unless pr p q).

Unfold u_everywhere u_unless.
Intros.
Apply u_sc_co.
Intros.
Unfold u_imp in H.
Unfold u_not u_and in H1.
Intros.
Elim H1.
Intros.
Generalize (H s H2).
Intro.
Generalize (H3 H4).
Intro.
Elim H5.
Qed.


Theorem u_stable_imp_unless : (pr:u_prog)(p,q:u_prop)
  (u_stable pr p)->(u_unless pr p q).

Unfold u_stable u_unless.
Intros.
Apply u_co_implication with 3:=H.
Unfold u_everywhere u_imp u_and u_not.
Tauto.
Unfold u_everywhere u_imp u_or.
Tauto.
Qed.

Theorem u_unless_imp_stable : (pr:u_prog)(p:u_prop)
  (u_unless pr p u_false)->(u_stable pr p).

Unfold u_unless u_stable.
Intros.
Apply u_co_implication with p':=(u_and p (u_not u_false)) q:=(u_or p
                                                              u_false).
Unfold u_everywhere u_imp u_and u_not u_false.
Intros.
Split.
Auto.
Unfold False.
Unfold not.
Auto.
Unfold u_everywhere u_imp u_or u_false.
Intros.
Elim H0.
Auto.
Intros.
Elim H1.
Auto.
Qed.


Theorem u_unless_big_disjunction : (pr:u_prog)(u_prog_not_empty pr)->
  (T:Set)(P,Q:T->u_prop)
  ((i:T)(u_unless pr (P i) (Q i)))->
  (u_unless pr (u_ex ? P) (u_ex ? Q)).

Intros.
Unfold u_unless.
Unfold u_unless in H0.
Generalize (u_co_big_disjunction T pr [i:T](u_and (P i) (u_not (Q i)))
             [i:T](u_or (P i) (Q i)) H0).
Intro.
Apply u_co_implication with p':=(u_ex T
                                  [i:T](u_and (P i) (u_not (Q i)))) q:=(
u_ex T [i:T](u_or (P i) (Q i))).
Unfold u_everywhere u_imp u_and u_ex u_not.
Intros.
Elim H2.
Intros.
Elim H3.
Intros t.
Intros.
Exists  t.
Split.
Assumption.
Unfold not.
Intros.
Unfold not in H4.
Apply H4.
Exists  t.
Assumption.
Unfold u_everywhere u_imp u_ex u_or.
Intros.
Elim H2.
Intros t.
Intros.
Elim H3.
Intros.
Left.
Exists  t.
Assumption.
Intros.
Right.
Exists  t.
Assumption.
Assumption.
Qed.


Theorem u_unless_implication_r : (pr:u_prog)(p,q,r:u_prop)
  (u_everywhere (u_imp q r)) ->
  (u_unless pr p q) -> (u_unless pr p r).

Unfold u_unless.
Intros.
Apply u_sc_co.
Intros.
Generalize (u_nc_co pr ? ? H0).
Intros.
Generalize (H3 t s H1).
Intros.
Clear  H3.
Clear  H1 H0.
Cut (u_or p q (u_effect t s)).
Unfold u_or.
Intros.
Elim H0.
Intros.
Left.
Assumption.
Intros.
Right.
Generalize (H (u_effect t s)).
Intros.
Apply H3.
Assumption.
Apply H4.
Unfold u_and u_not.
Unfold u_and u_not in H2.
Split.
Elim H2.
Intros.
Assumption.
Elim H2.
Intros.
Unfold not.
Intros.
Apply H1.
Apply (H s).
Assumption.
Qed.


Theorem u_unless_conjunction : (pr:u_prog)(p,p',q,q':u_prop)
  (u_unless pr p q) -> (u_unless pr p' q') -> 
  (u_unless pr (u_and p p') 
     (u_or3 (u_and p q') (u_and p' q) (u_and q q'))).

Intros.
Unfold u_unless in H.
Unfold u_unless in H0.
Unfold u_unless.
Apply u_sc_co.
Intros.
Generalize (u_nc_co pr (u_and p (u_not q)) (u_or p q) H).
Intros.
Generalize (u_nc_co pr (u_and p' (u_not q')) (u_or p' q') H0).
Intros.
Generalize (H3 t s).
Generalize (H3 t s H1).
Intros.
Generalize (H4 t s H1).
Intros.
Clear  H3.
Clear  H4.
Clear  H.
Clear  H0.
Clear  H1.
Elim H2.
Intros.
Clear  H2.
Clear  H6.
Cut (u_or p q (u_effect t s)).
Cut (u_or p' q' (u_effect t s)).
Intros.
Clear  H5.
Clear  H7.
Clear  H.
Clear  H0.
Unfold u_or u_or3 u_and.
Unfold u_or in H1.
Unfold u_or in H2.
Elim H1.
Elim H2.
Intros.
Left.
Auto.
Intros.
Right.
Right.
Left.
Auto.
Intros.
Elim H2.
Intros.
Right.
Left.
Auto.
Intros.
Right.
Right.
Right.
Auto.
Unfold u_and in H.
Unfold u_not u_or3 u_and in H0.
Apply H7.
Unfold u_and u_not.
Elim H.
Intros.
Split.
Assumption.
Unfold not.
Intros.
Elim H0.
Left.
Auto.
Unfold u_and in H.
Unfold u_not u_or3 u_and in H0.
Apply H5.
Unfold u_and u_not.
Elim H.
Intros.
Split.
Assumption.
Unfold not.
Intros.
Elim H0.
Right.
Left.
Auto.
Qed.


Theorem u_unless_simple_conjunction : (pr:u_prog)(p,p',q,q':u_prop)
  (u_unless pr p q) -> (u_unless pr p' q') -> 
  (u_unless pr (u_and p p') (u_or q q')).

Intros.
Apply u_unless_implication_r with q:=(u_or3 (u_and p q') (u_and p' q)
                                       (u_and q q')).
Unfold u_everywhere u_imp u_or3 u_and.
Intros.
Unfold u_or.
Elim H1.
Intros.
Elim H2.
Intros.
Right.
Assumption.
Intros.
Elim H2.
Intros.
Elim H3.
Intros.
Left.
Assumption.
Intros.
Left.
Elim H3.
Intros.
Assumption.
Apply u_unless_conjunction.
Assumption.
Assumption.
Qed.


Theorem u_unless_left_conjunction : (pr:u_prog)(p,p',q:u_prop)
  (u_unless pr p q) -> (u_unless pr p' q) -> 
  (u_unless pr (u_and p p') q).

Intros.
Apply u_unless_implication_r with q:=(u_or q q).
Unfold u_everywhere u_imp u_or.
Tauto.
Apply u_unless_simple_conjunction.
Trivial.
Trivial.
Qed.

Theorem u_unless_left_big_conjunction : (pr:u_prog)(u_prog_not_empty pr)->
  (b:nat)(Q:nat->u_prop)(r:u_prop)
  ((i:nat)(lt i b)->(u_unless pr (Q i) r))->
  (u_unless pr (u_all_nat b Q) r).

Intro pr.
Intro H.
Intro b.
Intro Q.
Generalize b.
Clear b.
Induction b.
Intros.
Clear H0.
Unfold u_unless.
Apply (u_co_implication_r pr (u_and (u_all_nat (0) Q) (u_not r)) u_true
        (u_or (u_all_nat (0) Q) r)).
Unfold u_everywhere u_imp u_true u_or u_all_nat.
Intros.
Left .
Intros.
Elim (lt_n_O ? H1).
Apply u_co_true.
Intros.
Cut (u_unless pr (Q n) r).
Intros.
Cut (i:nat)(lt i n)->(u_unless pr (Q i) r).
Intros.
Specialize (H0 r H3).
Intros.
Specialize (u_unless_left_conjunction pr ? ? ? H2 H4).
Intros.
Apply u_unless_subst_l with 2:=H5.
Unfold u_everywhere u_iff u_and u_all_nat.
Intros.
Split.
Intros.
Elim H6.
Intros.
Cut (le i n).
Intros.
Elim (le_lt_or_eq i n H10).
Intros.
Apply H9.
Trivial.
Intros.
Rewrite H11.
Trivial.
Apply lt_n_Sm_le.
Trivial.
Intros.
Split.
Apply H6.
Apply lt_n_Sn.
Intros.
Apply H6.
Apply lt_S.
Trivial.
Intros.
Apply H1.
Auto.
Apply H1.
Apply lt_n_Sn.
Qed.


Theorem u_unless_simple_disjunction : (pr:u_prog)(p,p',q,q':u_prop)
  (u_unless pr p q) -> (u_unless pr p' q') -> 
  (u_unless pr (u_or p p') (u_or q q')).

Intros.
Unfold u_unless in H.
Unfold u_unless in H0.
Unfold u_unless.
Generalize (u_co_disjunction pr ? ? ? ? H H0).
Intros.
Clear H H0.
Apply u_co_implication
 with p':=(u_or (u_and p (u_not q)) (u_and p' (u_not q')))
      q:=(u_or (u_or p q) (u_or p' q')).
Unfold u_everywhere u_imp u_and u_not u_or.
Intros.
Elim H.
Intros.
Clear H.
Elim H0.
Intros.
Left .
Split.
Trivial.
Unfold not.
Intros.
Apply H2.
Left .
Trivial.
Intros.
Right .
Split.
Trivial.
Unfold not.
Intros.
Apply H2.
Right .
Trivial.
Unfold u_everywhere u_imp u_or.
Intros.
Elim H.
Intros.
Elim H0.
Intros.
Left .
Left .
Trivial.
Intros.
Right .
Left .
Trivial.
Intros.
Elim H0.
Intros.
Left .
Right .
Trivial.
Intros.
Right .
Right .
Trivial.
Trivial.
Qed.


Theorem u_unless_disjunction : (pr:u_prog)(p,p',q,q':u_prop)
  (u_unless pr p q) -> (u_unless pr p' q') -> 
  (u_unless pr (u_or p p') 
     (u_or3 (u_and (u_not p) q') (u_and (u_not p') q) (u_and q q'))).

Intros pr p p' q q'.
Intros.
Unfold u_unless in H.
Unfold u_unless in H0.
Unfold u_unless.
Generalize (u_co_disjunction pr ? ? ? ? H H0).
Intros.
Clear H H0.
Apply u_co_implication
 with p':=(u_or (u_and p (u_not q)) (u_and p' (u_not q')))
      q:=(u_or (u_or p q) (u_or p' q')).
Unfold u_everywhere u_imp u_and u_not u_or u_or3.
Intros.
Elim H.
Clear H.
Intros.
Clear H1.
Generalize (classic (p s)).
Generalize (classic (p' s)).
Generalize (classic (q s)).
Generalize (classic (q' s)).
Intros.
Elim H.
Intros.
Elim H2.
Intros.
Elim H3.
Intros.
Elim H1.
Intros.
Elim H0.
Right .
Right .
Auto.

Intros.
Right .
Auto.

Intros.
Elim H0.
Right .
Left .
Auto.

Intros.
Left .
Auto.

Intros.
Elim H1.
Intros.
Elim H2.
Intros.
Elim H0.
Right .
Right .
Auto.

Intros.
Elim H4.
Intros.
Left .
Auto.

Intros.
Elim H0.
Left .
Auto.

Intros.
Right .
Auto.

Unfold u_everywhere u_imp u_and u_not u_or u_or3.
Intros.
Generalize (classic (p s)).
Generalize (classic (p' s)).
Generalize (classic (q s)).
Generalize (classic (q' s)).
Intros.
Clear H1.
Elim H.
Intros.
Elim H1.
Intros.
Left .
Auto.

Intros.
Elim H0.
Intros.
Right .
Right .
Right .
Auto.

Intros.
Elim H3.
Intros.
Left .
Auto.

Intros.
Right .
Right .
Left .
Auto.

Intros.
Elim H1.
Intros.
Left .
Auto.

Intros.
Elim H4.
Auto.

Auto.

Trivial.
Qed.


Theorem u_unless_cancelation : (pr:u_prog)(p,q,r:u_prop)
  (u_unless pr p q) -> (u_unless pr q r) -> (u_unless pr (u_or p q) r).

Intros.
Generalize (u_unless_disjunction pr p q q r).
Intros.
Apply (u_unless_implication_r pr (u_or p q)
        (u_or3 (u_and (u_not p) r) (u_and (u_not q) q) (u_and q r)) r).
Unfold u_everywhere.
Intros.
Unfold u_imp.
Unfold u_or3.
Unfold u_and.
Unfold u_not.
Intros.
Elim H2.
Intros.
Elim H3.
Intros.
Exact H5.
Intros.
Elim H3.
Intros.
Elim H4.
Unfold not.
Intros.
Elim H5.
Exact H6.
Intros.
Elim H4.
Intros.
Exact H6.
Apply H1.
Clear H1.
Exact H.
Clear H1.
Exact H0.
Qed.



Lemma u_elimination_1 : (x:u_var)(p:u_prop)(u_everywhere 
  (u_iff (u_ex ? [m:(u_var_type x)](u_and (u_with p x m) [s:u_st](s x)=m)) p)).

Intros.
Unfold u_everywhere u_iff.
Intros.
Unfold u_ex.
Split.
Intro.
Elim H.
Intros m.
Intro.
Unfold u_and in H0.
Elim H0.
Intros.
Unfold u_with in H1.
Rewrite <- H2 in H1.
Generalize (u_assign_unchanged s x).
Intro.
Rewrite H3 in H1.
Assumption.
Intro.
Exists  (s x).
Split.
Unfold u_with.
Rewrite -> (u_assign_unchanged s x).
Assumption.
Reflexivity.
Qed.


Theorem u_indep_imp_stable : (pr:u_prog)(p:u_prop)
  ((v:u_var)(u_independent p v))->(u_stable pr p).

Intros.
Unfold u_stable.
Apply u_sc_co.
Intros.
Apply u_fully_independent with p:=p s:=s.
Assumption.
Assumption.
Qed.



Lemma u_elimination_2 : (pr:u_prog)(x:u_var)(p:u_prop)
  ((v:u_var)~(v=x)->(u_independent p v)) ->
  (m:(u_var_type x))(u_stable pr (u_with p x m)).

Intros.
Apply u_indep_imp_stable.
Intro.
Elim (u_var_eq_dec v x).
Intros H1.
Rewrite -> H1.
Apply u_indep_with.
Intros H1.
Generalize (H v H1).
Intro.
Apply u_indep_imp_indep_with.
Assumption.
Qed.



Theorem u_elimination : 
  (pr:u_prog)(x:u_var)(p:u_prop)(q:(u_var_type x)->u_prop)
  ((v:u_var)~(v=x)->(u_independent p v)) ->
  ((m:(u_var_type x))(u_co pr [s:u_st]((s x)=m) (q m))) ->
  (u_co pr p (u_ex ? [m:(u_var_type x)](u_and (u_with p x m) (q m)))).

Intros.
Cut (m:(u_var_type x))
     (u_co pr (u_and (u_with p x m) [s:u_st](s x)=m)
       (u_and (u_with p x m) (q m))).
Intro.
Apply u_co_implication_l with (u_ex ?
                                [m:(u_var_type x)]
                                 (u_and (u_with p x m) [s:u_st](s x)=m)).
Apply u_iff_imp.
Apply u_iff_sym.
Apply u_elimination_1.
Apply u_co_big_disjunction.
Assumption.
Intro.
Apply u_co_conjunction.
Change (u_stable pr (u_with p x m)).
Apply u_elimination_2.
Assumption.
Apply H0.
Qed.


