(*C*
      Embedding UNITY into the Calculus of Inductive Constructions
                                     Using the COQ Proof Assistent
      Author:  Mark-Oliver Stehr
               University of Hamburg
               Department of Computer Science
               Theoretical Foundations of Computer Science Group 
               stehr@informatik.uni-hamburg.de
      Date:    September 1998
*C*)


Definition u_var_type := [v:u_var] (u_type (u_var_type_name v)).


Definition u_st := (v:u_var)(u_var_type v).



Definition u_val_eq :=
  [n:u_type_name][x:(u_type n)][n':u_type_name][y:(u_type n')]
  (eq_dep u_type_name u_type n x n' y).



Theorem u_eq_imp_val_eq:(n:u_type_name)
  (x,y:(u_type n))(x=y)->(u_val_eq n x n y).

Induction 1.
Unfold u_val_eq.
Constructor 1.
Qed.


Theorem u_val_eq_imp_eq:(n:u_type_name)
  (x,y:(u_type n))(u_val_eq n x n y)->(x=y).

Unfold u_val_eq.
Exact (eq_dep_eq u_type_name u_type).
Qed.


Definition u_cast := [n,n':u_type_name][e:(n=n')][x:(u_type n)]
  (eq_rec u_type_name n u_type x n' e).


Theorem u_cast_thm : (n:u_type_name)(e:(n=n))(x:(u_type n))
  ((u_cast n n e x)=x).

Intros.
Unfold u_cast.
Elim eq_rec_eq. 
Reflexivity.
Qed.

Lemma eq_lemma : 
  (T:Set)(P:T->T->Prop)((x:T)(P x x))->((x,y:T)(x=y)->(P x y)).

Intros.
Rewrite -> H0.
Apply H.
Qed.


Theorem u_cast_resp_eq:(n,n':u_type_name)(x:(u_type n))(e:(n=n'))
    (u_val_eq n' (u_cast n n' e x) n x).

Cut (n,n':u_type_name)
     (e:n=n')(x:(u_type n))(u_val_eq n' (u_cast n n' e x) n x).
Intro.
Intros.
Apply H.
Generalize (eq_lemma u_type_name
             [n,n':u_type_name]
              (e:n=n')(x:(u_type n))(u_val_eq n' (u_cast n n' e x) n x)).
Intros.
Apply H.
Intros.
Rewrite -> u_cast_thm.
Constructor 1.
Assumption.
Qed.


Axiom u_st_extensionality : (s,s':u_st)((v:u_var)((s v)=(s' v)))->(s=s').




Lemma u_var_eq_to_type_name_eq : (v,v':u_var)
  (v=v')->(u_var_type_name v)=(u_var_type_name v').

Intros.
Rewrite -> H.
Reflexivity.
Qed.


Definition u_assign := [s:u_st][v:u_var][x:(u_var_type v)]
  [v':u_var] <u_var_type v'>Case (u_var_eq_dec v v') of 
  (* v=v' *) [p:(v=v')] (u_cast (u_var_type_name v) (u_var_type_name v') 
                                (u_var_eq_to_type_name_eq v v' p) x) 
  (* ~(v=v') *) [_](s v')
  end.  




Theorem u_assign_unchanged : (s:u_st)(v:u_var)((u_assign s v (s v))=s).

Intros.
Apply u_st_extensionality.
Intros v'.
Unfold u_assign.
Elim (u_var_eq_dec v v').
Intros H.
Apply (u_val_eq_imp_eq (u_var_type_name v')).
Pattern 4 5 v' .
Rewrite <- H.
Apply u_cast_resp_eq.
Intro.
Reflexivity.
Qed.



Theorem u_assign_changed : (s:u_st)(v:u_var)(x:(u_var_type v))
  (((u_assign s v x) v)=x).

Unfold u_assign.
Intros.
Elim (u_var_eq_dec v v).
Intros H.
Apply (u_cast_thm (u_var_type_name v)).
Intro.
Unfold not in y.
Apply False_ind.
Apply y.
Reflexivity.
Qed.


Theorem u_assign_indep : (s:u_st)(v,v':u_var)(x:(u_var_type v))
  ~(v=v')->((u_assign s v x) v')=(s v').

Intros.
Unfold u_assign.
Elim (u_var_eq_dec v v').
Intros H1.
Apply False_ind.
Apply (H H1).
Intro.
Reflexivity.
Qed.


Theorem u_assign_twice : (s:u_st)(v:u_var)(x,y:(u_var_type v))
  (u_assign (u_assign s v x) v y)=(u_assign s v y).

Intros.
Apply u_st_extensionality.
Intros v'.
Elim (u_var_eq_dec v v').
Intros H.
Rewrite <- H.
Rewrite -> u_assign_changed.
Rewrite -> u_assign_changed.
Reflexivity.
Intros H.
Rewrite -> u_assign_indep.
Rewrite -> u_assign_indep.
Rewrite -> u_assign_indep.
Reflexivity.
Assumption.
Assumption.
Assumption.
Qed.


Theorem u_assign_commutes : (s:u_st)
  (v,v':u_var)(x:(u_var_type v))(y:(u_var_type v')) 
  ~(v=v')->(u_assign (u_assign s v x) v' y)=(u_assign (u_assign s v' y) v x).

Intros.
Apply u_st_extensionality.
Intros v''.
Elim (u_var_eq_dec v' v'').
Intros H1.
Rewrite <- H1.
Rewrite -> u_assign_changed.
Rewrite -> u_assign_indep.
Rewrite -> u_assign_changed.
Reflexivity.
Assumption.
Intro.
Rewrite -> u_assign_indep.
Elim (u_var_eq_dec v v'').
Intro.
Rewrite <- y1.
Rewrite -> u_assign_changed.
Rewrite -> u_assign_changed.
Reflexivity.
Intro.
Rewrite -> u_assign_indep.
Rewrite -> u_assign_indep.
Rewrite -> u_assign_indep.
Reflexivity.
Assumption.
Assumption.
Assumption.
Assumption.
Qed.


